"""Types for hub-related code."""
from dataclasses import dataclass
from typing import Dict, Final, Literal, TypedDict, Union

from typing_extensions import NotRequired

from bitfount.federated.monitoring.types import ProgressCounterDict
from bitfount.federated.types import SerializedProtocol
from bitfount.types import (
    _JSONDict,
    _S3PresignedPOSTFields,
    _S3PresignedPOSTURL,
    _S3PresignedURL,
)

# Hub/AM URLs
PRODUCTION_HUB_URL: Final[str] = "https://hub.bitfount.com"
PRODUCTION_AM_URL: Final[str] = "https://am.hub.bitfount.com"
_STAGING_HUB_URL: Final[str] = "https://hub.staging.bitfount.com"
_STAGING_AM_URL: Final[str] = "https://am.hub.staging.bitfount.com"
_DEV_HUB_URL: Final[str] = "http://localhost:3000"
_DEV_AM_URL: Final[str] = "http://localhost:3001"

# IDP URLs
_PRODUCTION_IDP_URL: Final[str] = (
    "https://prod-bitfount.eu.auth0.com/"
    "samlp/8iCJ33Kp6hc9ofrXTzr5GLxMRHWrlzZO?SAMLRequest="
)
_STAGING_IDP_URL: Final[str] = (
    "https://dev-bitfount.eu.auth0.com/"
    "samlp/Wk4XZHDKfY8F3OYcKdagIHETt6JYwX08?SAMLRequest="
)
_DEV_IDP_URL: Final[str] = (
    "https://dev-bitfount.eu.auth0.com/"
    "samlp/MP8oao6gcJd4jARwzJiJlEiK59ZeLCt3?SAMLRequest="
)


@dataclass
class APIKeys:
    """API keys for BitfountSession."""

    access_key_id: str
    access_key: str


# General JSONs
class _HubSuccessResponseJSON(TypedDict):
    """Generic hub success response JSON."""

    success: Literal[True]
    message: str


class _HubFailureResponseJSON(TypedDict):
    """Generic hub failure response JSON."""

    success: Literal[False]
    errorMessage: str


_HubResponseJSON = Union[_HubSuccessResponseJSON, _HubFailureResponseJSON]


# Hub-related JSON.
class _PodDetailsResponseJSON(TypedDict):
    """Response JSON from GET /api/pods/[userName]/[podName]."""

    podIdentifier: str
    podName: str
    podDisplayName: str
    podPublicKey: str
    accessManagerPublicKey: str
    description: str
    # dataSchema: str  # present but should not be used
    schemaStorageKey: str
    isOnline: bool
    providerUserName: str
    visibility: Literal["public", "private"]
    schemaDownloadUrl: _S3PresignedURL


class _MultiPodDetailsResponseJSON(TypedDict):
    """Response JSON from GET /api/pods."""

    podIdentifier: str
    name: str
    podDisplayName: str
    isOnline: bool
    podPublicKey: str
    accessManagerPublicKey: str
    description: str
    providerUserName: str
    podPagePath: str


class _PodRegistrationResponseJSON(TypedDict):
    """Response JSON from POST /api/pods."""

    success: bool
    alreadyExisted: bool
    message: str
    uploadUrl: _S3PresignedPOSTURL
    uploadFields: _S3PresignedPOSTFields


class _PodRegistrationFailureJSON(TypedDict):
    """Failure response JSON from POST /api/pods."""

    success: Literal[False]
    alreadyExisted: bool
    errorMessage: str


class _ModelDetailsResponseJSON(TypedDict):
    """Response JSON from GET /api/models when getting specific model."""

    modelStorageKey: str
    modelDownloadUrl: _S3PresignedURL
    modelHash: str
    modelVersion: NotRequired[int]


class _MultiModelDetailsResponseJSON(TypedDict):
    """Response JSON from GET /api/models when getting all models."""

    modellerName: str
    modelName: str
    modelStorageKey: str
    modelHash: str


class _ModelUploadResponseJSON(TypedDict):
    """Response JSON from POST /api/models."""

    # This is not a 1-to-1 mapping with the actual return types (i.e. some are
    # only returned on success, some only on failure), but is enough for our
    # use case.
    uploadUrl: _S3PresignedPOSTURL
    uploadFields: _S3PresignedPOSTFields
    success: bool
    errorMessage: str


class _MonitorPostJSON(TypedDict):
    """Form of the JSON object for sending to the task monitoring service."""

    taskId: str
    senderId: str
    recipientId: NotRequired[str]
    timestamp: str  # ISO 8601 format timestamp
    privacy: str  # one of monitoring.MonitorRecordPrivacy's values
    # one of _BitfountMessageType's names or monitoring.AdditionalMonitorMessageTypes
    type: str
    message: NotRequired[str]
    metadata: NotRequired[_JSONDict]
    progress: NotRequired[Dict[str, ProgressCounterDict]]
    resourceUsage: NotRequired[Dict[str, float]]


# Access Manager-related JSON
class _AccessManagerKeyResponseJSON(TypedDict):
    """Response JSON from GET /api/access-manager-key."""

    accessManagerPublicKey: str


class _SAMLChallengeResponseJSON(TypedDict):
    """Response JSON from GET /api/saml."""

    samlRequest: str
    id: str


class _SAMLAdditionalInfoPOSTJSON(TypedDict):
    """Required keys for JSON POST /api/saml."""

    # If this issue gets resolved then we can actually add the SAML response
    # fields to this as extras: https://github.com/python/mypy/issues/4617
    originalSAMLRequestID: str
    podIdentifier: str
    modellerName: str
    modellerProtocolRequest: SerializedProtocol
    identityProvider: Literal["SAML"]


class _OIDCAccessCheckPostJSON(TypedDict):
    """Required keys for OIDC JSON POST /api/access."""

    podIdentifier: str
    modellerProtocolRequest: SerializedProtocol
    modellerName: str
    modellerToken: str
    identityProvider: Literal["OIDC"]


class _SignatureBasedAccessCheckPostJSON(TypedDict):
    """Required keys for Signatured based JSON POST /api/access."""

    podIdentifier: str
    modellerName: str
    modellerRequest: SerializedProtocol
    unsignedTask: bytes
    signedTask: bytes
    identityProvider: Literal["SIGNATURE"]


class _AMAccessCheckResponseJSON(TypedDict):
    """Response JSON from Access Manager access check.

    Covers:
        - POST /api/access
    """

    code: Literal[
        # Common response types
        "ACCEPT",
        # /api/access response types
        "NO_ACCESS",
        "INVALID_PROOF_OF_IDENTITY",
        "UNAUTHORISED",
        "NO_PROOF_OF_IDENTITY",
    ]


# Auth0 related types
class _DeviceCodeRequestDict(TypedDict):
    """Data dictionary for POST request to /oauth/device/code.

    See: https://auth0.com/docs/api/authentication?http#device-authorization-flow
    """

    audience: str
    scope: str
    client_id: str


class _DeviceCodeResponseJSON(TypedDict):
    """JSON response for POST /oauth/device/code.

    See: https://auth0.com/docs/api/authentication?http#device-authorization-flow
    """

    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: str
    expires_in: int
    interval: int


class _DeviceAccessTokenRequestDict(TypedDict):
    """Data dictionary for device code POST request to /oauth/token.

    See: https://auth0.com/docs/api/authentication?http#device-authorization-flow48
    """

    grant_type: Literal["urn:ietf:params:oauth:grant-type:device_code"]
    client_id: str
    device_code: str


class _DeviceAccessTokenResponseJSON(TypedDict):
    """Success JSON response for POST /oauth/token.

    For Device Authorization Flow.

    See: https://auth0.com/docs/api/authentication?http#device-authorization-flow48
    """

    access_token: str
    id_token: str
    refresh_token: str
    scope: str
    expires_in: int
    token_type: Literal["Bearer"]


class _TokenRefreshRequestDict(TypedDict):
    """Data dictionary for token refresh POST request to /oauth/token.

    This is not the full potential params, but is enough for us.

    See: https://auth0.com/docs/api/authentication?http#refresh-token
    """

    grant_type: Literal["refresh_token"]
    client_id: str
    refresh_token: str


class _TokenRefreshResponseJSON(TypedDict):
    """Success JSON response for refresh token POST /oauth/token.

    See: https://auth0.com/docs/api/authentication?http#refresh-token

    Note that our response will include a new refresh token as we are using
    refresh token rotation.

    See: https://auth0.com/docs/secure/tokens/refresh-tokens/refresh-token-rotation
    """

    access_token: str
    id_token: str
    refresh_token: str  # see docstring
    scope: str
    expires_in: int
    token_type: Literal["Bearer"]


class _DeviceAccessTokenFailResponseJSON(TypedDict):
    """Fail JSON response for POST /oauth/token.

    For Device Authorization Flow.

    See: https://auth0.com/docs/api/authentication?http#device-authorization-flow48
    """

    error: str
    error_description: str


class _PKCEAccessTokenRequestDict(TypedDict):
    """Data dictionary for ACF with PKCE code POST request to /oauth/token.

    See: https://auth0.com/docs/api/authentication?http#authorization-code-flow-with-pkce45  # noqa: B950
    """

    grant_type: Literal["authorization_code"]
    client_id: str
    code: str
    code_verifier: str
    redirect_uri: str


class _PKCEAccessTokenResponseJSON(TypedDict):
    """Success JSON response for POST /oauth/token.

    For Authorization Code Flow with PKCE.

    See: https://auth0.com/docs/api/authentication?http#authorization-code-flow-with-pkce45  # noqa: B950
    """

    access_token: str
    refresh_token: str
    id_token: str
    token_type: Literal["Bearer"]
    expires_in: int



"""Column averaging algorithm."""
from __future__ import annotations

from typing import Any, ClassVar, Dict, Mapping, Optional

from marshmallow import fields
import numpy as np

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.algorithms.base import (
    _BaseAlgorithmFactory,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.types import T_FIELDS_DICT
from bitfount.utils import delegates

logger = _get_federated_logger(__name__)


class _ModellerSide(_BaseModellerAlgorithm):
    """Modeller side of the ColumnAverage algorithm."""

    def initialise(
        self,
        **kwargs: Any,
    ) -> None:
        """Nothing to initialise here."""
        pass

    def run(
        self, results: Mapping[str, Dict[str, float]]
    ) -> Dict[str, Dict[str, float]]:
        """Simply returns results."""
        return dict(results)


class _WorkerSide(_BaseWorkerAlgorithm):
    """Worker side of the ColumnAverage algorithm."""

    def __init__(self, *, field: str, table_name: str, **kwargs: Any) -> None:
        self.datasource: BaseSource
        self.field = field
        self.table_name = table_name
        super().__init__(**kwargs)

    def initialise(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        **kwargs: Any,
    ) -> None:
        """Sets Datasource."""
        self.datasource = datasource

    def run(self) -> Dict[str, float]:
        """Returns the mean of the field in `BaseSource` dataframe."""
        mean = np.mean(
            self.datasource.get_column(col_name=self.field, table_name=self.table_name)
        )
        return {"mean": float(mean)}


@delegates()
class ColumnAverage(_BaseAlgorithmFactory):
    """Simple algorithm for taking the arithmetic mean of a column in a table.

    Args:
        field: The name of the column to take the mean of.
        table_name: The name of the table on which column
            average will be performed on.

    Attributes:
        field: The name of the column to take the mean of.
        table_name: The name of the table on which column
            average will be performed on.
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "field": fields.Str(),
        "table_name": fields.Str(),
    }

    def __init__(self, *, field: str, table_name: str, **kwargs: Any):
        super().__init__(**kwargs)
        self.field = field
        self.table_name = table_name

    def modeller(self, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the ColumnAverage algorithm."""
        return _ModellerSide(**kwargs)

    def worker(self, **kwargs: Any) -> _WorkerSide:
        """Returns the worker side of the ColumnAverage algorithm."""
        return _WorkerSide(field=self.field, table_name=self.table_name, **kwargs)



"""Tests for the column averaging algorithm."""
from typing import TYPE_CHECKING
from unittest.mock import Mock

import numpy as np
from pytest_mock import MockerFixture

from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.utils import DatabaseConnection
from bitfount.federated.algorithms.base import (
    _BaseAlgorithm,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.algorithms.column_avg import (
    ColumnAverage,
    _ModellerSide,
    _WorkerSide,
)
from bitfount.federated.utils import _ALGORITHMS
from bitfount.schemas.utils import bf_dump, bf_load
from tests.utils.helper import create_dataset, unit_test


class TestColumnAverage:
    """Tests ColumnAverage algorithm."""

    @unit_test
    def test_modeller(self) -> None:
        """Tests modeller method."""
        algorithm_factory = ColumnAverage(field="TARGET", table_name="fake")
        algorithm = algorithm_factory.modeller()
        for type_ in [_BaseAlgorithm, _BaseModellerAlgorithm]:
            assert isinstance(algorithm, type_)

    @unit_test
    def test_worker(self) -> None:
        """Tests worker method."""
        algorithm_factory = ColumnAverage(field="TARGET", table_name="fake")
        algorithm = algorithm_factory.worker()
        for type_ in [_BaseAlgorithm, _BaseWorkerAlgorithm]:
            assert isinstance(algorithm, type_)

    @unit_test
    def test_worker_run_single_table_dataframe_datasource(self) -> None:
        """Tests worker run method with a single table datasource."""
        datasource = DataFrameSource(create_dataset())
        algorithm_factory = ColumnAverage(field="TARGET", table_name="doesnt_matter")
        worker = algorithm_factory.worker()
        worker.initialise(datasource=datasource)
        result = worker.run()
        assert isinstance(result, dict)
        assert isinstance(result["mean"], float)

    @unit_test
    def test_worker_run_single_table_database_datasource(
        self, mock_engine: Mock, mocker: MockerFixture
    ) -> None:
        """Tests worker run method with a single table datasource."""
        data = create_dataset()
        db_conn = DatabaseConnection(mock_engine, table_names=["dummy_data"])
        algorithm_factory = ColumnAverage(field="TARGET", table_name="doesnt_matter")
        # Mocks `Session` and resulting context manager
        mock_session = Mock()
        mock_result = Mock()
        mock_session.execute.return_value = mock_result
        mock_result.columns.return_value = [
            (i,) for i in data[algorithm_factory.field].to_list()
        ]
        mocker.patch(
            "bitfount.data.datasources.database_source.Session"
        ).return_value.__enter__.return_value = mock_session

        datasource = DatabaseSource(db_conn)
        worker = algorithm_factory.worker()
        worker.initialise(datasource=datasource)
        result = worker.run()
        assert isinstance(result, dict)
        assert isinstance(result["mean"], float)
        assert result["mean"] == np.mean(data[algorithm_factory.field])


@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for column average algorithm."""

    def test_serialization(self) -> None:
        """Test Marshmallow Serialization for column average algorithm."""
        algorithm_factory = ColumnAverage(field="TARGET", table_name="fake")
        dumped = bf_dump(algorithm_factory)
        loaded = bf_load(dumped, _ALGORITHMS)
        assert algorithm_factory.__dict__ == loaded.__dict__


# Static tests for algorithm-protocol compatibility
if TYPE_CHECKING:
    from typing import cast

    from bitfount.federated.protocols.results_only import (
        _ResultsOnlyCompatibleAlgoFactory_,
        _ResultsOnlyCompatibleModellerAlgorithm,
        _ResultsOnlyDataIncompatibleWorkerAlgorithm,
    )

    # Check compatible with ResultsOnly
    _algo_factory: _ResultsOnlyCompatibleAlgoFactory_ = ColumnAverage(
        field=cast(str, object()), table_name=cast(str, object())
    )
    _modeller_side: _ResultsOnlyCompatibleModellerAlgorithm = _ModellerSide()
    _worker_side: _ResultsOnlyDataIncompatibleWorkerAlgorithm = _WorkerSide(
        field=cast(str, object()), table_name=cast(str, object())
    )



"""Tests types.py."""
from datetime import datetime
import string
from typing import Literal, Union

import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd
import pytest

from bitfount.data.types import _ExtendableLabelEncoder
from tests.utils.helper import create_dataset, unit_test


@unit_test
class TestExtendableLabelEncoder:
    """Tests the extendable label encoder."""

    DATA = np.array(list(string.ascii_lowercase))
    DATA2 = np.array(list(str(x) for x in range(0, 10)))

    def test_add_values(self) -> None:
        """Checks we can successfully add values.

        Also checks we don't get extras if doing it twice.
        """
        len1 = len(self.DATA)
        len2 = len(self.DATA2)
        encoder = _ExtendableLabelEncoder()
        encoder.add_values(self.DATA)
        assert encoder.size == len1
        encoder.add_values(self.DATA)
        assert encoder.size == len1
        encoder.add_values(self.DATA2)
        assert encoder.size == len1 + len2

    def test_transform_str(self) -> None:
        """Checks we can encode a column that is strings."""
        encoder = _ExtendableLabelEncoder()
        encoder.add_values(self.DATA)
        encoder.add_values(self.DATA2)
        input_arr = pd.Series(["a", "0", "2", "b"])
        encoded = encoder.transform(input_arr)
        assert_array_equal(encoded, np.array([0, 26, 28, 1]))

    def test_transform_error(self) -> None:
        """Tests transform error."""
        encoder = _ExtendableLabelEncoder()
        encoder.add_values(self.DATA)
        encoder.add_values(self.DATA2)
        with pytest.raises(ValueError):
            encoder.transform(pd.Series(["a", "a", "0", "2", "unseen"]))

    def test_size(self) -> None:
        """Tests size property explicitly."""
        encoder = _ExtendableLabelEncoder()
        encoder.classes = dict(zip(["a", "b", "c"], [1, 2, 3]))
        assert encoder.size == 3

    def test_eq_magic_method(self) -> None:
        """Tests that __eq__ works as expected."""
        encoder1 = _ExtendableLabelEncoder()
        encoder1.add_values(self.DATA)
        encoder2 = _ExtendableLabelEncoder()
        encoder2.add_values(self.DATA2)
        encoder3 = _ExtendableLabelEncoder()
        encoder3.add_values(self.DATA)
        assert not (encoder1 == encoder2)
        assert encoder1 == encoder3

    @pytest.mark.parametrize("values_type", ["pandas", "numpy"])
    def test_date_column(
        self, values_type: Union[Literal["pandas"], Literal["numpy"]]
    ) -> None:
        """Tests that encoding with a date column works."""
        dataframe = create_dataset()
        date_column = dataframe["Date"]
        encoder = _ExtendableLabelEncoder()
        if values_type == "pandas":
            encoder.add_values(date_column)
        else:
            # Adds just the unique values as a numpy array
            encoder.add_values(np.asarray(date_column.unique()))
        transformed_column = encoder.transform(date_column)
        assert sorted(set(transformed_column)) == sorted(encoder.classes.values())

    @pytest.mark.parametrize("values_type", ["pandas", "numpy"])
    def test_datetime_column(
        self, values_type: Union[Literal["pandas"], Literal["numpy"]]
    ) -> None:
        """Tests that encoding with a datetime column works."""
        dataframe = create_dataset()
        dataframe["Date"] = datetime(2020, 1, 1, 11, 30, 15)
        date_column = dataframe["Date"]
        encoder = _ExtendableLabelEncoder()
        if values_type == "pandas":
            encoder.add_values(date_column)
        else:
            # Adds just the unique values as a numpy array
            encoder.add_values(np.asarray(date_column.unique()))
        transformed_column = encoder.transform(date_column)
        assert sorted(set(transformed_column)) == sorted(encoder.classes.values())

    @pytest.mark.parametrize("values_type", ["pandas", "numpy"])
    def test_datetime_column_with_empty_time(
        self, values_type: Union[Literal["pandas"], Literal["numpy"]]
    ) -> None:
        """Tests that encoding with a datetime column with no time works."""
        dataframe = create_dataset()
        dataframe["Date"] = datetime(2020, 1, 1)
        date_column = dataframe["Date"]
        encoder = _ExtendableLabelEncoder()
        if values_type == "pandas":
            encoder.add_values(date_column)
        else:
            # Adds just the unique values as a numpy array
            encoder.add_values(np.asarray(date_column.unique()))
        transformed_column = encoder.transform(date_column)
        assert sorted(set(transformed_column)) == sorted(encoder.classes.values())



"""Tests early_stopping.py module."""
from typing import Callable, Dict, Generator, List

from _pytest.logging import LogCaptureFixture
import pytest
from pytest import fixture

from bitfount.federated.early_stopping import FederatedEarlyStopping
from tests.utils.helper import unit_test

ValidationMetrics = List[Dict[str, float]]
ValidationMetricsGenerator = Generator[ValidationMetrics, None, None]
ValidationMetricsCreator = Callable[..., Generator[List[Dict[str, float]], None, None]]


@unit_test
class TestFederatedEarlyStopping:
    """Tests FederatedEarlyStopping class."""

    @fixture
    def decreasing_validation_metrics(self) -> ValidationMetricsCreator:
        """Validation metrics that should trigger early stopping when iterated."""

        def validation_metrics_generator() -> ValidationMetricsGenerator:
            validation_metrics = [{"AUC": 0.8}]
            for _ in range(10):
                validation_metrics.append({"AUC": validation_metrics[-1]["AUC"] - 0.01})
                yield validation_metrics

        return validation_metrics_generator

    @fixture
    def increasing_validation_metrics(self) -> ValidationMetricsCreator:
        """Validation metrics that should not trigger early stopping when iterated."""

        def validation_metrics_generator() -> ValidationMetricsGenerator:
            validation_metrics = [{"AUC": 0.8}]
            for _ in range(10):
                validation_metrics.append({"AUC": validation_metrics[-1]["AUC"] + 0.01})
                yield validation_metrics

        return validation_metrics_generator

    @fixture
    def rollercoaster_validation_metrics(self) -> ValidationMetricsCreator:
        """Validation metrics that first go down and then go up."""

        def validation_metrics_generator() -> ValidationMetricsGenerator:
            validation_metrics = [{"AUC": 0.8}]
            for _ in range(5):
                validation_metrics.append({"AUC": validation_metrics[-1]["AUC"] - 0.01})
                yield validation_metrics

            for _ in range(5):
                validation_metrics.append({"AUC": validation_metrics[-1]["AUC"] + 0.01})
                yield validation_metrics

        return validation_metrics_generator

    @pytest.mark.parametrize("patience", [1, 3, 7])
    def test_early_stopping_patience(
        self, decreasing_validation_metrics: ValidationMetricsCreator, patience: int
    ) -> None:
        """Tests that early stopping is triggered at the right time."""
        early_stopping = FederatedEarlyStopping(
            metric="AUC", patience=patience, delta=0.01
        )

        for epoch, metrics in enumerate(decreasing_validation_metrics(), start=1):
            stop = early_stopping.check(metrics)
            if epoch > patience:
                assert stop
            else:
                assert not stop

    @pytest.mark.parametrize("delta", [0.001, 0.00999, 0.01, 0.02])
    def test_early_stopping_delta(
        self, delta: float, increasing_validation_metrics: ValidationMetricsCreator
    ) -> None:
        """Tests that early stopping is triggered at the right time."""
        patience = 5
        early_stopping = FederatedEarlyStopping(
            metric="AUC", patience=patience, delta=delta
        )

        for epoch, metrics in enumerate(increasing_validation_metrics(), start=1):
            stop = early_stopping.check(metrics)
            if epoch > patience and delta > 0.01:  # fixture increments AUC by 0.01
                assert stop
            else:
                assert not stop

    @pytest.mark.parametrize("patience", [1, 3, 7])
    def test_early_stopping_not_triggered(
        self, increasing_validation_metrics: ValidationMetricsCreator, patience: int
    ) -> None:
        """Tests that early stopping is not triggered at the right time."""
        early_stopping = FederatedEarlyStopping(
            metric="AUC", patience=patience, delta=0.01
        )

        for _, metrics in enumerate(increasing_validation_metrics(), start=1):
            stop = early_stopping.check(metrics)
            assert not stop

    @pytest.mark.parametrize("patience", [1, 3, 7])
    def test_early_stopping_with_variable_metrics(
        self, patience: int, rollercoaster_validation_metrics: ValidationMetricsCreator
    ) -> None:
        """Tests early stopping with metrics that go down and then up."""
        delta = 0.01
        metric = "AUC"
        halfway_point_index = 5
        early_stopping = FederatedEarlyStopping(
            metric=metric, patience=patience, delta=delta
        )
        for epoch, metrics in enumerate(rollercoaster_validation_metrics(), start=1):
            stop = early_stopping.check(metrics)

            # Make assertions about counter
            if epoch <= halfway_point_index:
                assert early_stopping.counter == epoch
            elif epoch < patience:
                assert early_stopping.counter == halfway_point_index
            elif metrics[-1][metric] - metrics[-1 - patience][metric] >= delta:
                assert early_stopping.counter == 0
            else:
                assert early_stopping.counter == halfway_point_index

            # Make assertions about stop
            if early_stopping.counter > patience:
                assert stop
            else:
                assert not stop

    @pytest.mark.parametrize("patience", [1, 3, 7])
    def test_early_stopping_metric_missing(
        self,
        caplog: LogCaptureFixture,
        decreasing_validation_metrics: ValidationMetricsCreator,
        patience: int,
    ) -> None:
        """Checks early stopping ignored and warning issued if missing metric."""
        early_stopping = FederatedEarlyStopping(
            metric="missing_metric", patience=patience, delta=0.01
        )

        for _, metrics in enumerate(decreasing_validation_metrics(), start=1):
            stop = early_stopping.check(metrics)
            assert not stop
            assert (
                caplog.records[-1].msg
                == "Early stopping ignored. "
                + "Metric missing_metric not reported by model."
            )
            assert caplog.records[-1].levelname == "WARNING"



"""Exceptions related to monitoring modules."""
from bitfount.exceptions import BitfountError


class MonitorModuleError(BitfountError):
    """Base exception for monitor module errors."""

    pass


class NoMonitorModuleError(MonitorModuleError):
    """Error for when a monitor module should exist but doesn't."""

    pass


class ExistingMonitorModuleError(MonitorModuleError):
    """Error for when a monitor module already exists.

    Raised when trying to create a second monitor module.
    """

    pass



"""Contains pytest fixtures related to hub and access manager interactions."""
import inspect
from typing import cast
from unittest.mock import Mock

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from pytest import fixture
from pytest_mock import MockerFixture


@fixture
def authoriser_public_key() -> RSAPublicKey:
    """Authoriser public key fixture."""
    serialized = inspect.cleandoc(
        """-----BEGIN PUBLIC KEY-----
        MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAifKZtWWu9JqavRu1pZOS
        / 9NKyWJ7gvTCibVeoXpiSvRaVVnaQ+VIQveMWr3O5871CKLB1DtDjpLw1hyvOIg6
        FTzt2odicizRUML / RAHzKEbZ1fWaCt9imjtsESrJN2rumjmtvO96CU8BLaIAx7Kv
        khd1HWeMKxQzekTjo4AU80T4CJMWSKzliNFxo6YFEP8FVwQmV7Q0C4F8mV9Qv0hu
        bV96wY2H0DsPDRXbUcoETBjFZG25+JxrsaV3fMxJmK / 3OBO8JQ0O9yvfAk8wCGLB
        coIEgiWGcbDS6x3uzbutYaONvgzaGm5Ysj8HjfujaJbptVjXB / 6+KkMANHWtFcva
        wQIDAQAB
        -----END PUBLIC KEY-----"""
    )
    return cast(
        RSAPublicKey,
        serialization.load_pem_public_key(
            serialized.encode(), backend=default_backend()
        ),
    )


@fixture
def access_manager_private_key() -> RSAPrivateKey:
    """Access manager private key fixture."""
    serialized = inspect.cleandoc(
        """-----BEGIN RSA PRIVATE KEY-----
        MIIEogIBAAKCAQEAyXmGr0cJ3ZfBrllSI1YQKuDp5UIa+zn3PksveKea4fbqQhZ2
        0mBMJNGDxgUCucF8R9PbbsY801qcLTBlZgc4H/OJd8PE1Wv+gxPH5akYxb9v7ZoZ
        H5In6Vv9UwxoJP3yxHegcnx5ayIpfAvIqtYMM01gdNxUKMpxWyNhnDIG7qHzZTnK
        JS3mmOVJxT9F8QNVzG39v3SsjstQDKAGunA63JEl0jtsqBg4Z34TQ/vVkwtaVmA4
        x6UTQgEOMtgftp607r8FrZt/4rYuI1WUiwrS2enqDc8ALRKCRzWmvJcqKfKmS0JG
        /TOQKsUfFGAhOpqodI1VyhMPhFaH+RMdOIZRfQIDAQABAoIBAA7Uey17tSYyfXcv
        6/Pd6eK+G3lcnE/RvHlBX4UqESSA4E1tCeICZJhxM5nn1IOH0nYXChflT6TfDo/k
        +BHChkcTzhKavmpXQKy1UAl/oOstZDfBGAhS4c0zkDoEt8XBS5VamNzUfLt+ED+p
        PmQUzWEZWPhBH5DBFgb3W7/7jrcR4zXaDL2G9Jmr2B6YZbHsGNJevfLx2HB5n4Kp
        DhiIT+NPUNVeMRwcQVvJbfZzPpxc0+Km+0xCi2ggIIPbGv4LdCryxreydCRgstnh
        TRQE5y3rl8UGX7dn311EFyxeLsPg6lCbjbzC2u/HgAiiiuARxaoWN2qgykZVbs8y
        3fevLgECgYEA70Vlda9EnzGV4ayqld9940H+9uYwuqWsTG2Nn9anlRpNzpBqN9oZ
        mQSz5pLxUejdXV8BP1uBc6cJsGca72u4UYtI0wdINHvRnO8TH0MkTrOacpiOSBQv
        f5poi6ZyFHUQZhxQP5n9i/WCCW65M72DYJYPnyvbylzp0gb12HYzQM0CgYEA14+h
        P0F/B7BO4xWU4ndRFTjxxPRydxgi1ffFhPK6MNEqEqgvHHvF3jdWmxRR/mD5wY7H
        8gh8Xu7WpHu7jTTn+BBOoHmvpDGRZcOI/NpDUlQDmIona19iA60K2ogQHanc09gG
        61nW70P95NE+Tbi0e3VsOi6MxJBNcjKOWQB1k3ECgYAWYkSanNnrlqTx5shSmwTY
        5MyoZmJ6o5MZ10nfgSJOVZaf02zMas/oO8Nkt6xxwDotSP0B41qZ6i+C3DJt4YLt
        xfz7GPxNRKQYbVk5nHtiDXVOcS0mWxUSd+x93AViGo19L/rl6E8K793Jah1TVNBc
        kc1GU8ENDmIlbtdC99PVpQKBgBhc3F2Y2nGVBKV3t/cMgzaB1Bltg05ghPdn5IQN
        JuwDh9zJ82ElwAxIdtpaJpY4gaHVsFVzjhLwrYOzlFP2Sc1okmpitm45R/SXb/Eb
        /gXdBvcDjkb9ZceClDp0rFWd8B5lSQrqgXdTWOu2lnMUUBmwAMZmonRI1uf0PNBk
        U+WRAoGAal4aCAJjW63QbLS7yipDP63vMqKE1Pg5mkksQ7NHP9W7KHsB9IQO9JZQ
        5BKyFv6dY7Dk5jHqa9J/tO/YHI5C2/hOe/pXBiarHNSE6MR/ItK6X30oW5rYj/aO
        OhEdX5Mr37Fa9UKGdkBZ7VbpYlIAnMCccuus4cZG3az/5BQTEy0=
        -----END RSA PRIVATE KEY-----"""
    )
    return cast(
        RSAPrivateKey,
        serialization.load_pem_private_key(
            serialized.encode(), backend=default_backend(), password=None
        ),
    )


@fixture
def access_manager_public_key() -> RSAPublicKey:
    """Access manager public key fixture."""
    serialized = inspect.cleandoc(
        """-----BEGIN RSA PUBLIC KEY-----
        MIIBCgKCAQEAyXmGr0cJ3ZfBrllSI1YQKuDp5UIa+zn3PksveKea4fbqQhZ20mBM
        JNGDxgUCucF8R9PbbsY801qcLTBlZgc4H/OJd8PE1Wv+gxPH5akYxb9v7ZoZH5In
        6Vv9UwxoJP3yxHegcnx5ayIpfAvIqtYMM01gdNxUKMpxWyNhnDIG7qHzZTnKJS3m
        mOVJxT9F8QNVzG39v3SsjstQDKAGunA63JEl0jtsqBg4Z34TQ/vVkwtaVmA4x6UT
        QgEOMtgftp607r8FrZt/4rYuI1WUiwrS2enqDc8ALRKCRzWmvJcqKfKmS0JG/TOQ
        KsUfFGAhOpqodI1VyhMPhFaH+RMdOIZRfQIDAQAB
        -----END RSA PUBLIC KEY-----"""
    )
    return cast(
        RSAPublicKey,
        serialization.load_pem_public_key(
            serialized.encode(), backend=default_backend()
        ),
    )


@fixture
def modeller_public_key_string() -> str:
    """Modeller public key as a string fixture."""
    return inspect.cleandoc(
        """
        -----BEGIN PUBLIC KEY-----
        MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAruIdWcWQKIph+mQd6XVk
        nbs95a0mslUq7Y6ukeCgE1nwqphTMDYefbvCXNQ9C1CX/b7ydyoZBSQ92RdsR19t
        J6AzpfZco3DflTamqxiqxyy75uMNni/hos6sCU0MR82IbvIe9IrA+ZXWsWDuJXsy
        mHNLvJ+nhD4w26y4+cu4ONxo9vW0+hFHzC2yg8RS9AsjwQ/NzmUvw2nGkbR4FfZZ
        GUONQhvqKr96nDBsWx2nmIjO3uDnI0pK9JLiT30M2Gma+QK5WQI7zYZYkXZRY4rf
        8Lsc5DO9Tc6MLEaNv46m332rVMNtl5o0TPuHLFIkN6hg58z4WRmZvSLswG6972qY
        NwIDAQAB
        -----END PUBLIC KEY-----"""
    )


@fixture
def modeller_public_key(modeller_public_key_string: str) -> RSAPublicKey:
    """Modeller public key fixture."""
    return cast(
        RSAPublicKey,
        serialization.load_pem_public_key(
            modeller_public_key_string.encode(), backend=default_backend()
        ),
    )


@fixture
def modeller_private_key() -> RSAPrivateKey:
    """Modeller private key fixture."""
    serialized = inspect.cleandoc(
        """
        -----BEGIN RSA PRIVATE KEY-----
        MIIEpAIBAAKCAQEAruIdWcWQKIph+mQd6XVknbs95a0mslUq7Y6ukeCgE1nwqphT
        MDYefbvCXNQ9C1CX / b7ydyoZBSQ92RdsR19tJ6AzpfZco3DflTamqxiqxyy75uMN
        ni / hos6sCU0MR82IbvIe9IrA+ZXWsWDuJXsymHNLvJ+nhD4w26y4+cu4ONxo9vW0
        +hFHzC2yg8RS9AsjwQ / NzmUvw2nGkbR4FfZZGUONQhvqKr96nDBsWx2nmIjO3uDn
        I0pK9JLiT30M2Gma+QK5WQI7zYZYkXZRY4rf8Lsc5DO9Tc6MLEaNv46m332rVMNt
        l5o0TPuHLFIkN6hg58z4WRmZvSLswG6972qYNwIDAQABAoIBAFjnwK4Tnq7NcvPd
        E8vMj0MUnmn1tbdxTqqIH70vaZGM5qmDWL8otCWNX0jb6NCovywTd6YW87NGcl67
        5qL+IzrC2Z7keMHqLu / nYaGXn7Izzq6Y / SJPCzE+Sf1MexLBrTthwVTmeSYpA4UX
        VtodIGKaMoWI0JuNFj+HNl5fL59PrfcPoDeo5rmm161fhIfaROvXBl526nowRZwC
        Z1Z / DmJfYOerO07m6Mx4LQwPOM5+WxqorN+rL+ / ARgLBF3K8ehYqHAjyskq73UzG
        33fBqlB9YfyYrDHSfdY5tQFCxreg5mrLI7rH0mtDct03QSR7EXFfc+Xp2875m8qi
        +vAeo4kCgYEA5fsWZZcKdwJxeNZoKweoHFEvGEQ7SQ4QYY3U3tlffVXrMNXFQ7XG
        RgQjw2wtDEz15yGYDezxfIFoeIACi2Kf3ZGNi697Jq68FYcKClz0sNX94ArfqivT
        wUrZEKmAFB5xZ52TzMdJs0B8G7dN2Mo9x8LadGZ4iJY8WgSTDq43b6sCgYEAwqs9
        9OKfT / M5JTqh1zu7HhKTAGoLGPGtJjDFApT1Df0RQtIOG6QYqEsM9plbJJ8kJ4+F
        kbZTxNh7cMGhsPxK / WW0ToCPVUDb06L1sSggEZEeHGaLE6TMVNZKLBl6VpZMYptO
        GRSkVxJegGPndGguoIRxJEwlmzz5Ed4VWMnb3aUCgYEAoBeSdrOW2FN+FAUEzHdv
        Ag9NflH5BxjgNFicB2k // KqzpvIYeNLvq8uz7ZS / IyHrx7Jt+7umOuqGg+NwYgEM
        khuBcex9COd4DHBNf / tHZlYAfVZ0bixGIvbsdZnYi6jJfryFp3ZPCpXoDw8iBLh8
        GzO8CG1QmJwkdPFcQUrrUjMCgYA2dW1pupRS63oLHjH0YnNgYFXbgc8E9Wc5Dd+v
        bzh251sS1Gy58tgrDIPd4N0Mi7CzmIUHVHhu0xkFXEX9mzbnWLJuW6y9bt2QgUA2
        hUEjaPOBgnZZ0nqPKTuqxp3g5z8LPuNERPAMo8joJgq0GAkjQqncO6kxmocbJoPW
        SPJHOQKBgQC7bspk7qkMthCSdpnvgBpeJI+FLgQkh46jw5b2NEoVs5VzOpYklQxV
        ddeMJgGRrmp2lzMU8oVlhtaElgCW / 11U4SbTPrktxFQWKP4gpVZ / qM6OFitUMv88
        D9yDvPTnj / Wyc2lpbhJF1GL8bzj / ygQfACjne2IGMJovOgc1hcxdHw ==
        -----END RSA PRIVATE KEY-----"""
    )
    return cast(
        RSAPrivateKey,
        serialization.load_pem_private_key(
            serialized.encode(), password=None, backend=default_backend()
        ),
    )


@fixture
def modeller_message() -> str:
    """Example modeller message."""
    return "TEST Message"


@fixture
def bitfount_model_correct_structure() -> str:
    """Example good BitfountModel."""
    return inspect.cleandoc(
        """
        from typing import Optional
        from bitfount.models.bitfount_model import BitfountModel
        from bitfount.federated.mixins import _DistributedModelMixIn

        class MyModel(BitfountModel, _DistributedModelMixIn):

            def __init__(self, **kwargs) -> None:
                super().__init__(**kwargs)
                self.epochs: Optional[int] = 1
                self.steps: Optional[int] = None
                self._total_num_batches_trained: int = 0

            @staticmethod
            def backend_tensor_shim():
                ...

            def tensor_precision(self):
                ...

            def get_param_states(self):
                ...

            def apply_weight_updates(self):
                ...

            def update_params(self):
                ...

            def deserialize_params(self):
                ...

            def serialize_params(self):
                ...

            def diff_params(self):
                ...

            def set_model_training_iterations(self):
                ...

            def reset_trainer(self) -> None:
                ...

            def fit(self):
                ...

            def log_(self):
                ...

            def initialise_model(self):
                ...

            def evaluate(self):
                ...

            def serialize(self):
                ...

            def deserialize(self):
                ...

            def _fit_local(self):
                ...

            @staticmethod
            def _get_import_statements():
                ...

            def _get_model():
                ...
        """
    )


@fixture
def bitfount_model_incorrect_structure() -> str:
    """Example BitfountModel which doesn't implement DistributedModelProtocol.

    Some required methods from `bitfount_model_correct_structure` have been
    commented out.
    """
    return inspect.cleandoc(
        """
        from bitfount.models.bitfount_model import BitfountModel

        class MyModel(BitfountModel):

            def __init__(self, **kwargs) -> None:
                super().__init__(**kwargs)
                self.epochs: Optional[int] = 1
                self.steps: Optional[int] = None
                self._total_num_batches_trained: int = 0

            @staticmethod
            def backend_tensor_shim():
                ...

            def tensor_precision(self):
                ...

            def get_param_states(self):
                ...

            def apply_weight_updates(self):
                ...

            # def update_params(self):
            #     ...

            # def diff_params(self):
            #     ...

            def set_model_training_iterations(self):
                ...

            def reset_trainer(self) -> None:
                ...

            def fit(self):
                ...

            def log_(self):
                ...

            def initialise_model(self):
                ...

            def evaluate(self):
                ...

            def serialize(self):
                ...

            def deserialize(self):
                ...

            def _fit_local(self):
                ...

            @staticmethod
            def _get_import_statements():
                ...

            def _get_model():
                ...
        """
    )


@fixture
def bitfount_model_correct_structure_size(bitfount_model_correct_structure: str) -> int:
    """Size of the correct structure BitfountModel fixture in storage."""
    return len(bitfount_model_correct_structure.encode("utf-8"))


@fixture
def mock_s3_data_upload_in_api_module(mocker: MockerFixture) -> Mock:
    """Mocks out the S3 data upload call in api.py.

    This call is used to upload the schema.
    """
    mock_s3_upload: Mock = mocker.patch(
        "bitfount.hub.api._upload_data_to_s3", autospec=True
    )
    return mock_s3_upload


@fixture
def mock_s3_file_upload_in_api_module(mocker: MockerFixture) -> Mock:
    """Mocks out the S3 file upload call in api.py.

    This call is used to upload models.
    """
    mock_s3_upload: Mock = mocker.patch(
        "bitfount.hub.api._upload_file_to_s3", autospec=True
    )
    return mock_s3_upload


@fixture
def mock_s3_data_download_in_api_module(mocker: MockerFixture) -> Mock:
    """Mocks out the S3 data download call in api.py.

    This call is used to download the schema.
    """
    mock_s3_download: Mock = mocker.patch(
        "bitfount.hub.api._download_data_from_s3", autospec=True
    )
    return mock_s3_download


@fixture
def mock_s3_file_download_in_api_module(mocker: MockerFixture) -> Mock:
    """Mocks out the S3 file download call in api.py.

    This call is used to download models.
    """
    mock_s3_download: Mock = mocker.patch(
        "bitfount.hub.api._download_file_from_s3", autospec=True
    )
    return mock_s3_download



"""Tests for binary transformations."""
from typing import Type

from marshmallow import ValidationError
import pytest

from bitfount.transformations.base_transformation import Transformation
from bitfount.transformations.binary_operations import (
    AdditionTransformation,
    BinaryOperation,
    ComparisonTransformation,
    DivisionTransformation,
    MultiplicationTransformation,
    NumericBinaryOperation,
    SubtractionTransformation,
)
from tests.bitfount.transformations.transformation_test_helpers import (
    gen_name_fails,
    gen_name_test,
    registration_test,
)
from tests.utils.helper import unit_test


@unit_test
class TestBinaryOperation:
    """Tests for the BinaryOperation abstract transformation."""

    def test__gen_name_fails(self) -> None:
        """Tests that name generation fails."""
        gen_name_fails(BinaryOperation)


@unit_test
class TestNumericBinaryOperation:
    """Tests for the NumericBinaryOperation abstract transformation."""

    def test__gen_name_fails(self) -> None:
        """Tests that name generation fails."""
        gen_name_fails(NumericBinaryOperation)

    def test_schema_loads(self) -> None:
        """Tests schema loads correctly."""
        s = NumericBinaryOperation.schema()
        t = s.loads(
            """
            name: test
            output: True
            arg1: hello
            arg2: world
            """
        )
        assert t.name == "test"
        assert t.output is True
        assert t.arg1 == "hello"
        assert t.arg2 == "world"

    def test_schema_loads_arg2_numeric(self) -> None:
        """Tests schema loads if arg2 is a number."""
        s = NumericBinaryOperation.schema()
        t = s.loads(
            """
            name: test
            output: True
            arg1: hello
            arg2: 2
            """
        )
        assert t.name == "test"
        assert t.output is True
        assert t.arg1 == "hello"
        assert t.arg2 == 2.0

    def test_schema_load_fails_arg1_nonstring(self) -> None:
        """Tests schema load fails if arg1 isn't a string."""
        s = NumericBinaryOperation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: True
                arg1: 1
                arg2: world
                """
            )

    def test_schema_load_fails_arg2_wrong_type(self) -> None:
        """Tests schema load fails if arg2 isn't a float or str."""
        s = NumericBinaryOperation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: True
                arg1: hello
                arg2: []
                """
            )

    def test_schema_load_fails_missing_arg1(self) -> None:
        """Tests schema load fails if missing arg1."""
        s = NumericBinaryOperation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: True
                arg2: world
                """
            )

    def test_schema_load_fails_missing_arg2(self) -> None:
        """Tests schema load fails if missing arg2."""
        s = NumericBinaryOperation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: True
                arg1: hello
                """
            )


def gen_name_test_binary(cls: Type[Transformation], registry_name: str) -> None:
    """Tests name generation for binary transformations.

    Test that a binary transformation class is created with the correctly generated
    random name.
    """
    gen_name_test(cls, registry_name, arg1="hello", arg2="world")


@unit_test
class TestAdditionTransformation:
    """Tests for the AdditionTransformation class."""

    def test__gen_name(self, fake_uuid: None) -> None:
        """Tests transformation name generation."""
        gen_name_test_binary(AdditionTransformation, "add")

    def test_registration(self) -> None:
        """Tests transformation registration."""
        registration_test(AdditionTransformation, "add")


@unit_test
class TestSubtractionTransformation:
    """Tests for the SubtractionTransformation class."""

    def test__gen_name(self, fake_uuid: None) -> None:
        """Tests transformation name generation."""
        gen_name_test_binary(SubtractionTransformation, "subtract")

    def test_registration(self) -> None:
        """Tests transformation registration."""
        registration_test(SubtractionTransformation, "subtract")


@unit_test
class TestMultiplicationTransformation:
    """Tests for the MultiplicationTransformation class."""

    def test__gen_name(self, fake_uuid: None) -> None:
        """Tests transformation name generation."""
        gen_name_test_binary(MultiplicationTransformation, "multiply")

    def test_registration(self) -> None:
        """Tests transformations registration."""
        registration_test(MultiplicationTransformation, "multiply")


@unit_test
class TestDivisionTransformation:
    """Tests for the DivisionTransformation class."""

    def test__gen_name(self, fake_uuid: None) -> None:
        """Tests transformation name generation."""
        gen_name_test_binary(DivisionTransformation, "divide")

    def test_registration(self) -> None:
        """Tests transformation registration."""
        registration_test(DivisionTransformation, "divide")


@unit_test
class TestComparisonTransformation:
    """Tests for the ComparisonTransformation class."""

    def test__gen_name(self, fake_uuid: None) -> None:
        """Tests transformation name generation."""
        gen_name_test_binary(ComparisonTransformation, "compare")

    def test_registration(self) -> None:
        """Tests transformation registration."""
        registration_test(ComparisonTransformation, "compare")



"""Pod communication protocols.

These classes take an algorithm and are responsible for organising the communication
between Pods and Modeller.

Attributes:
    registry: A read-only dictionary of protocol factory names to their
        implementation classes.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
import inspect
from pathlib import Path
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Collection,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
)

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
import pandas as pd

from bitfount.data.datasources.base_source import BaseSource
import bitfount.federated.algorithms.base as algorithms
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModelAlgorithmFactory,
)
from bitfount.federated.authorisation_checkers import IdentityVerificationMethod
from bitfount.federated.helper import (
    _check_and_update_pod_ids,
    _create_message_service,
    _get_idp_url,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.modeller import _Modeller
from bitfount.federated.pod_vitals import _PodVitals
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.federated.roles import _RolesMixIn
from bitfount.federated.transport.base_transport import _BaseMailbox
from bitfount.federated.transport.config import MessageServiceConfig
from bitfount.federated.transport.message_service import _MessageService
from bitfount.federated.transport.modeller_transport import _ModellerMailbox
from bitfount.federated.transport.worker_transport import _WorkerMailbox
from bitfount.federated.types import ProtocolType, SerializedProtocol
from bitfount.hub.helper import _default_bitfounthub
from bitfount.types import T_FIELDS_DICT, T_NESTED_FIELDS, _BaseSerializableObjectMixIn

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub
    from bitfount.hub.authentication_flow import BitfountSession


logger = _get_federated_logger(__name__)

MB = TypeVar("MB", bound=_BaseMailbox)


class _BaseProtocol(Generic[MB], ABC):
    """Blueprint for modeller side or the worker side of BaseProtocolFactory."""

    def __init__(
        self,
        *,
        algorithm: Union[_BaseCompatibleModeller, _BaseCompatibleWorker],
        mailbox: MB,
        **kwargs: Any,
    ):

        self.algorithm = algorithm
        self.mailbox = mailbox

        super().__init__(**kwargs)


class _BaseCompatibleModeller(Protocol):
    """Protocol defining base modeller-side compatibility."""

    pass


class _BaseModellerProtocol(_BaseProtocol[_ModellerMailbox], ABC):
    """Modeller side of the protocol.

    Calls the modeller side of the algorithm.
    """

    def __init__(
        self,
        *,
        algorithm: _BaseCompatibleModeller,
        mailbox: _ModellerMailbox,
        **kwargs: Any,
    ):
        super().__init__(algorithm=algorithm, mailbox=mailbox, **kwargs)

    @abstractmethod
    async def run(
        self,
        **kwargs: Any,
    ) -> Union[List[Any], pd.DataFrame]:
        """Runs Modeller side of the protocol."""
        pass


class _BaseCompatibleWorker(Protocol):
    """Protocol defining base worker-side compatibility."""

    pass


class _BaseWorkerProtocol(_BaseProtocol[_WorkerMailbox], ABC):
    """Worker side of the protocol.

    Calls the worker side of the algorithm.
    """

    def __init__(
        self,
        *,
        algorithm: _BaseCompatibleWorker,
        mailbox: _WorkerMailbox,
        **kwargs: Any,
    ):
        super().__init__(algorithm=algorithm, mailbox=mailbox, **kwargs)

    @abstractmethod
    async def run(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        pod_vitals: Optional[_PodVitals] = None,
        pod_identifier: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Runs the worker-side of the algorithm."""
        pass


# The mutable underlying dict that holds the registry information
_registry: Dict[str, Type[_BaseProtocolFactory]] = {}
# The read-only version of the registry that is allowed to be imported
registry: Mapping[str, Type[_BaseProtocolFactory]] = MappingProxyType(_registry)


class _BaseCompatibleAlgoFactory(Protocol):
    """Protocol defining base algorithm factory compatibility."""

    class_name: str
    fields_dict: ClassVar[T_FIELDS_DICT] = {}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {}


class _BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
    """Base Protocol from which all other protocols must inherit."""

    fields_dict: ClassVar[T_FIELDS_DICT] = {}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {"algorithm": algorithms.registry}

    def __init__(
        self,
        *,
        algorithm: _BaseCompatibleAlgoFactory,
        **kwargs: Any,
    ) -> None:
        self.class_name = ProtocolType[type(self).__name__].value
        super().__init__(**kwargs)
        self._validate_algorithm(algorithm)
        self.algorithm = algorithm

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to Protocol registry")
            _registry[cls.__name__] = cls

    @classmethod
    @abstractmethod
    def _validate_algorithm(cls, algorithm: _BaseCompatibleAlgoFactory) -> None:
        """Checks that `algorithm` is compatible with the protocol.

        Raises TypeError if `algorithm` is not compatible with the protocol.
        """
        pass

    @abstractmethod
    def modeller(
        self, mailbox: _ModellerMailbox, **kwargs: Any
    ) -> _BaseModellerProtocol:
        """Creates an instance of the modeller-side for this protocol."""
        raise NotImplementedError

    @abstractmethod
    def worker(
        self, mailbox: _WorkerMailbox, hub: BitfountHub, **kwargs: Any
    ) -> _BaseWorkerProtocol:
        """Creates an instance of the worker-side for this protocol."""
        raise NotImplementedError

    @abstractmethod
    def dump(self) -> SerializedProtocol:
        """Serialize object, returns as JSON dict."""
        raise NotImplementedError

    def run(
        self,
        pod_identifiers: Collection[str],
        session: Optional[BitfountSession] = None,
        username: Optional[str] = None,
        hub: Optional[BitfountHub] = None,
        ms_config: Optional[MessageServiceConfig] = None,
        message_service: Optional[_MessageService] = None,
        pod_public_key_paths: Optional[Mapping[str, Path]] = None,
        identity_verification_method: IdentityVerificationMethod = IdentityVerificationMethod.DEFAULT,  # noqa: B950
        private_key_or_file: Optional[Union[RSAPrivateKey, Path]] = None,
        idp_url: Optional[str] = None,
        require_all_pods: bool = False,
    ) -> Optional[Any]:
        """Sets up a local Modeller instance and runs the protocol.

        Args:
            pod_identifiers: The BitfountHub pod identifiers to run against.
            session: Optional. Session to use for authenticated requests.
                 Created if needed.
            username: Username to run as. Defaults to logged in user.
            hub: BitfountHub instance. Default: hub.bitfount.com.
            ms_config: Message service config. Default: messaging.bitfount.com.
            message_service: Message service instance, created from ms_config if not
                provided. Defaults to "messaging.bitfount.com".
            pod_public_key_paths: Public keys of pods to be checked against.
            identity_verification_method: The identity verification method to use.
            private_key_or_file: Private key (to be removed).
            idp_url: The IDP URL.
            require_all_pods: If true raise PodResponseError if at least one pod
                identifier specified rejects or fails to respond to a task request.

        Returns:
            Results of the protocol.

        Raises:
            PodResponseError: If require_all_pods is true and at least one pod
                identifier specified rejects or fails to respond to a task request.
            ValueError: If attempting to train on multiple pods, and the
                `DataStructure` table name is given as a string.
        """
        hub = _default_bitfounthub(hub=hub, username=username)

        if isinstance(self.algorithm, _BaseModelAlgorithmFactory):
            if (
                len(pod_identifiers) > 1
                and hasattr(self.algorithm.model.datastructure, "table")
                and isinstance(self.algorithm.model.datastructure.table, str)
            ):
                raise ValueError(
                    "You are attempting to train on multiple pods, and the"
                    "provided the DataStructure table name is a string. "
                    "Please make sure that the `table` argument to the "
                    "`DataStructure` is a mapping of Pod names to table names. "
                )
            pod_identifiers = _check_and_update_pod_ids(pod_identifiers, hub)
            datastructure_pod_identifiers = (
                self.algorithm.model.datastructure.get_pod_identifiers()
            )
            if datastructure_pod_identifiers:
                datastructure_pod_identifiers = _check_and_update_pod_ids(
                    datastructure_pod_identifiers, hub
                )
                self.algorithm.model.datastructure._update_datastructure_with_hub_identifiers(
                    datastructure_pod_identifiers
                )
        if not session:
            session = hub.session
        if not idp_url:
            idp_url = _get_idp_url()
        if not message_service:
            message_service = _create_message_service(
                session=session,
                ms_config=ms_config,
            )

        modeller = _Modeller(
            protocol=self,
            message_service=message_service,
            bitfounthub=hub,
            pod_public_key_paths=pod_public_key_paths,
            identity_verification_method=identity_verification_method,
            private_key=private_key_or_file,
            idp_url=idp_url,
        )
        name = type(self).__name__

        logger.info(f"Starting {name} Job...")

        result = modeller.run(pod_identifiers, require_all_pods=require_all_pods)
        logger.info(f"Completed {name} Job...")
        return result



"""Typed wrappers around the Bitfount REST api."""
from typing import List

from bitfount.hub.api import BitfountAM, BitfountHub, PodPublicMetadata
from bitfount.hub.authentication_flow import BitfountSession
from bitfount.hub.exceptions import (
    ModelTooLargeError,
    ModelUploadError,
    ModelValidationError,
    PodDoesNotExistError,
    SchemaUploadError,
)
from bitfount.hub.helper import get_pod_schema
from bitfount.hub.types import PRODUCTION_AM_URL, PRODUCTION_HUB_URL, APIKeys

__all__: List[str] = [
    "APIKeys",
    "BitfountAM",
    "BitfountHub",
    "BitfountSession",
    "ModelTooLargeError",
    "ModelUploadError",
    "ModelValidationError",
    "PRODUCTION_AM_URL",
    "PRODUCTION_HUB_URL",
    "PodDoesNotExistError",
    "PodPublicMetadata",
    "SchemaUploadError",
    "get_pod_schema",
]

# See top level `__init__.py` for an explanation
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False



"""Tests for the message service implementations."""
import asyncio
import logging
import sys
from typing import Callable
from unittest.mock import Mock

import pytest
from pytest import fixture

from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
from tests.utils.helper import (
    AUC_THRESHOLD,
    create_local_modeller_and_workers,
    integration_test,
)

logging.basicConfig(level=logging.INFO, stream=sys.stdout)


@integration_test
class TestMessageServiceUsage:
    """Tests for Message Service federated transport layer for training jobs."""

    # TODO: [BIT-983] Add tests to check the following:
    #           - Other models
    #           - Other algorithms
    #           - Other protocols
    #           - Transformations
    #           - Early stopping

    @fixture
    def mock_get_pod_public_keys(
        self, apply_mock_get_pod_public_keys: Callable[[str], Mock]
    ) -> Mock:
        """Mocks out get_pod_public_keys function in modeller.py."""
        return apply_mock_get_pod_public_keys(
            "bitfount.federated.modeller._get_pod_public_keys"
        )

    @pytest.mark.skip(
        "[BIT-1914] LGBMRandomForestClassifier incompatible with ResultsOnly"
    )
    async def test_classifier_runs(
        self,
        mock_get_pod_public_keys: Mock,
        mock_message_aes_decryption: Mock,
        mock_message_aes_encryption: Mock,
        mock_rsa_encryption: Mock,
        mock_rsa_sign_message: Mock,
    ) -> None:
        """Tests that a Logistic Regression classifier runs."""
        modeller, workers = create_local_modeller_and_workers(
            model_name="LogisticRegressionClassifier",
            protocol_name="ResultsOnly",
            algorithm_name="ModelTrainingAndEvaluation",
        )

        modeller_results, *_ = await asyncio.gather(
            modeller.run_async(
                [worker.mailbox.pod_identifier for worker in workers], True
            ),
            *[_run_func_and_listen_to_mailbox(w.run(), w.mailbox) for w in workers]
        )
        for result in modeller_results:
            assert result is not None
            assert isinstance(result, dict)
            auc = result["AUC"]
            assert auc > AUC_THRESHOLD



"""Dealing with interactions with configuration and environment variables."""
from functools import lru_cache
import logging
import os
from pathlib import Path
from typing import Callable, Dict, Final, List, Optional, Tuple

import GPUtil
from environs import Env

__all__: List[str] = [
    "BITFOUNT_DEV_MODE",
    "BITFOUNT_ENGINE",
    "BITFOUNT_HOME",
    "BITFOUNT_KEY_STORE",
    "BITFOUNT_LOGS_DIR",
    "BITFOUNT_LOG_TO_FILE",
    "BITFOUNT_ONLINE_CHECK_HARD_LIMIT",
    "BITFOUNT_ONLINE_CHECK_SOFT_LIMIT",
    "BITFOUNT_PLUGIN_PATH",
    "BITFOUNT_STORAGE_PATH",
    "BITFOUNT_TB_LIMIT",
    "get_gpu_metadata",
]

logger = logging.getLogger(__name__)
env = Env()

# Public Config/Environment Variables
BITFOUNT_HOME: Path = Path(os.getenv("BITFOUNT_HOME", Path.home()))
BITFOUNT_STORAGE_PATH: Path = BITFOUNT_HOME / ".bitfount"
BITFOUNT_PLUGIN_PATH: Path = BITFOUNT_STORAGE_PATH / "_plugins"
BITFOUNT_KEY_STORE: Path = BITFOUNT_STORAGE_PATH / "known_workers.yml"
BITFOUNT_LOGS_DIR: Path = Path("bitfount_logs")

# Private Config/Environment Variables
_BITFOUNT_CLI_MODE: bool = False
_PRODUCTION_ENVIRONMENT: Final[str] = "production"
_STAGING_ENVIRONMENT: Final[str] = "staging"
_DEVELOPMENT_ENVIRONMENT: Final[str] = "dev"
_ENVIRONMENT_CANDIDATES: Tuple[str, ...] = (
    _PRODUCTION_ENVIRONMENT,
    _STAGING_ENVIRONMENT,
    _DEVELOPMENT_ENVIRONMENT,
)


@lru_cache(maxsize=1)
def _get_environment() -> str:
    """Returns bitfount environment to be used from BITFOUNT_ENVIRONMENT variable.

    The result is cached to avoid multiple warning messages. This means that changes to
    the `BITFOUNT_ENVIRONMENT` environment variable will not be detected whilst the
    library is running.

    Returns:
        str: PRODUCTION_ENVIRONMENT, STAGING_ENVIRONMENT or DEVELOPMENT_ENVIRONMENT

    """
    BITFOUNT_ENVIRONMENT = os.getenv("BITFOUNT_ENVIRONMENT", _PRODUCTION_ENVIRONMENT)
    if BITFOUNT_ENVIRONMENT not in _ENVIRONMENT_CANDIDATES:
        raise ValueError(
            f"The environment specified by the environment variable "
            f"BITFOUNT_ENVIRONMENT ({BITFOUNT_ENVIRONMENT}) is not in the supported "
            f"list of environments ({_ENVIRONMENT_CANDIDATES})"
        )
    if BITFOUNT_ENVIRONMENT == _STAGING_ENVIRONMENT:
        logger.warning(
            "Using the staging environment. "
            + "This will only work for Bitfount employees."
        )
    if BITFOUNT_ENVIRONMENT == _DEVELOPMENT_ENVIRONMENT:
        logger.warning(
            "Using the development environment. "
            + "This will only work if you have all Bitfount services running locally."
        )

    return BITFOUNT_ENVIRONMENT


_PYTORCH_ENGINE: Final[str] = "pytorch"
_BASIC_ENGINE: Final[str] = "basic"
_ENGINE_CANDIDATES: Tuple[str, ...] = (
    _BASIC_ENGINE,
    _PYTORCH_ENGINE,
)

# Set BITFOUNT_ENGINE, defaulting to PYTORCH_ENGINE or BASIC_ENGINE
# Start with BASIC_ENGINE as default
BITFOUNT_ENGINE: str = _BASIC_ENGINE
try:
    # Use the type specified by envvar if present
    BITFOUNT_ENGINE = os.environ["BITFOUNT_ENGINE"]
    # Check that the engine option is a valid one
    if BITFOUNT_ENGINE not in _ENGINE_CANDIDATES:
        raise ValueError(
            f"The backend engine specified by the environment variable "
            f"BITFOUNT_ENGINE ({BITFOUNT_ENGINE}) is not in the supported list of "
            f"backends ({_ENGINE_CANDIDATES})"
        )
except KeyError:
    # Otherwise, if PyTorch is installed use PYTORCH_ENGINE
    try:
        import torch  # noqa: F401

        BITFOUNT_ENGINE = _PYTORCH_ENGINE
    except ImportError:
        pass


def _get_gpu_metadata_gputil() -> Tuple[Optional[str], int]:
    """Returns gpu metadata from GPUtil.

    Uses the name of the first GPU thereby assuming that there is only 1 type of GPU
    attached to the machine.

    Returns:
        Tuple[Optional[str], int]: name of gpu and how many there are
    """
    gpus = GPUtil.getGPUs()
    if gpus:
        return gpus[0].name, len(gpus)
    # nvidia-smi installed, but no GPU available
    return None, 0


def _get_gpu_metadata_pytorch() -> Tuple[Optional[str], int]:
    """Return gpu metadata from pytorch.

    Returns:
        Tuple[Optional[str], int]: name of gpu and how many there are
    """
    import torch.cuda as cuda

    if cuda.is_available():
        # Devices without CUDA can still use pytorch,
        # but this will throw an exception
        return cuda.get_device_name(), cuda.device_count()
    else:
        raise Exception("CUDA not available.")


_GPU_COUNT_FUNCTION_LOOKUP: Dict[str, Callable[..., Tuple[Optional[str], int]]] = {
    _BASIC_ENGINE: _get_gpu_metadata_gputil,
    _PYTORCH_ENGINE: _get_gpu_metadata_pytorch,
}


def get_gpu_metadata() -> Tuple[Optional[str], int]:
    """Retrieve details about GPUs if available.

    Uses tools available in the appropriate backend,
    to find GPUs that are usable by the backend.

    Returns: a tuple of GPU name and count.
    """
    # noinspection PyBroadException
    try:
        return _GPU_COUNT_FUNCTION_LOOKUP[BITFOUNT_ENGINE]()
    except Exception as ex:
        # Broad exception handling here as libraries may throw various exceptions
        # But if anything is raised we can assume we don't have GPU access
        logger.warning(f"Encountered exception whilst gathering GPU information: {ex}")
        logger.warning("No GPU info will be used.")
        return None, 0


# Message Service Environment Variables
BITFOUNT_ONLINE_CHECK_SOFT_LIMIT: int = env.int(
    "BITFOUNT_ONLINE_CHECK_SOFT_LIMIT", default=180, validate=lambda n: n > 0
)
BITFOUNT_ONLINE_CHECK_HARD_LIMIT: int = env.int(
    "BITFOUNT_ONLINE_CHECK_HARD_LIMIT", default=180, validate=lambda n: n > 0
)


# Logging/Error Handling Config/Environment Variables
BITFOUNT_DEV_MODE: bool = env.bool("BITFOUNT_DEV_MODE", default=False)
BITFOUNT_LOG_TO_FILE: bool = env.bool("BITFOUNT_LOG_TO_FILE", default=True)
BITFOUNT_TB_LIMIT: int = env.int("BITFOUNT_TB_LIMIT", default=1)



"""Implements loss for pytorch modules."""

# Based on https://github.com/justusschock/dl-utils/blob/master/dlutils/losses/soft_dice.py # noqa: B950
from typing import Optional, Sequence, Union, cast

import numpy as np
import torch
import torch.nn.functional as F


class SoftDiceLoss(torch.nn.Module):
    """Soft Dice Loss.

    The soft dice loss is computed as a fraction of nominator over denominator, where:
    nominator is 2 * the area of overlap between targets and predictions plus a
    smooth factor,and the denominator is the total number of pixels in both images
    plus the smooth factor.If weights are provided the fraction is multiplied by the
    provided weights for each class.If either square_nom or square_denom are provided,
    then the respective nominator or denominator will be raised to the power of 2.

    Args:
        square_nom: Whether to square the nominator. Optional.
        square_denom: Whether to square the denominator. Optional.
        weight: Additional weighting of individual classes. Optional.
        smooth: Smoothing for nominator and denominator. Optional.Defaults to 1.
    """

    def __init__(
        self,
        square_nom: bool = False,
        square_denom: bool = False,
        weight: Optional[Union[Sequence, torch.Tensor]] = None,
        smooth: float = 1.0,
    ):
        super().__init__()
        self.square_nom = square_nom
        self.square_denom = square_denom

        self.smooth = smooth

        if weight is not None:
            if not isinstance(weight, torch.Tensor):
                weight = torch.tensor(weight)

            self.register_buffer("weight", weight)
        else:
            self.weight = None

    def forward(self, predictions: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """Computes Soft Dice Loss.

        Args:
            predictions: The predictions obtained by the network.
            targets: The targets (ground truth) for the predictions.

        Returns:
            torch.Tensor: The computed loss value

        Raises:
            ValueError: If the predictions tensor has less than 3 dimensions.
            ValueError: If the targets tensor has less than 2 dimensions.

        """
        # number of classes for onehot
        n_classes = predictions.shape[1]
        if predictions.ndim < 3:
            raise ValueError(
                "Predictions expected to have at least 3 dimensions. "
                f"Your predictions have {predictions.ndim}"
            )
        if targets.ndim < 2:
            raise ValueError(
                "Targets expected to have at least 2 dimensions. "
                f"Your targets have {targets.ndim}"
            )

        # one hot encoding on the batch
        with torch.no_grad():
            target = targets.unsqueeze(1)
            dtype, device, shape = target.dtype, target.device, target.shape
            one_hot = torch.zeros(
                shape[0], n_classes, *shape[2:], dtype=dtype, device=device
            )
            targets_onehot = one_hot.scatter_(1, target, 1.0)

        # sum over spatial dimensions
        dims = tuple(range(2, predictions.dim()))

        # compute nominator
        if self.square_nom:
            nom = torch.sum((predictions * targets_onehot.float()) ** 2, dim=dims)
        else:
            nom = torch.sum(predictions * targets_onehot.float(), dim=dims)
        nom = 2 * nom + self.smooth

        # compute denominator
        if self.square_denom:
            i_sum = torch.sum(predictions**2, dim=dims)
            t_sum = torch.sum(targets_onehot**2, dim=dims)
        else:
            i_sum = torch.sum(predictions, dim=dims)
            t_sum = torch.sum(targets_onehot, dim=dims)

        denom = i_sum + t_sum.float() + self.smooth

        # compute loss
        frac = nom / denom

        # apply weight for individual classes properly
        if self.weight is not None:
            frac = self.weight * frac  # type: ignore[unreachable] # Reason: weight can be a torch.tensor. # noqa: B950
        # average over classes
        frac = -torch.mean(frac, dim=1)

        return frac


def soft_dice_loss(
    pred: np.ndarray,
    targets: np.ndarray,
    square_nom: bool = False,
    square_denom: bool = False,
    weight: Optional[Union[Sequence, torch.Tensor]] = None,
    smooth: float = 1.0,
) -> torch.Tensor:
    """Functional implementation of the SoftDiceLoss.

    Args:
        pred: A numpy array of predictions.
        targets: A numpy array of targets.
        square_nom: Whether to square the nominator.
        square_denom: Whether to square the denominator.
        weight: Additional weighting of individual classes.
        smooth: Smoothing for nominator and denominator.

    Returns:
        A torch tensor with the computed dice loss.
    """
    dice_loss = SoftDiceLoss(square_nom, square_denom, weight, smooth)
    # Make predictions and targets tensors
    pred_tensor, targets_tensor = torch.from_numpy(pred), torch.from_numpy(targets)
    # Targets need to be of torch type long.
    targets_tensor = targets_tensor.long()
    pred_tensor = F.softmax(pred_tensor, dim=1)
    return cast(torch.Tensor, dice_loss(pred_tensor, targets_tensor))



"""Modules related to the transformations framework.

All transformations should be imported here so that they are automatically added to the
transformations registry.
"""
from typing import List

from bitfount.transformations.base_transformation import TRANSFORMATION_REGISTRY
from bitfount.transformations.batch_operations import (
    IMAGE_TRANSFORMATIONS,
    AlbumentationsImageTransformation,
)
from bitfount.transformations.binary_operations import (
    AdditionTransformation,
    ComparisonTransformation,
    DivisionTransformation,
    MultiplicationTransformation,
    SubtractionTransformation,
)
from bitfount.transformations.dataset_operations import (
    CleanDataTransformation,
    NormalizeDataTransformation,
)
from bitfount.transformations.exceptions import (
    IncorrectReferenceError,
    InvalidBatchTransformationError,
    MissingColumnReferenceError,
    NotColumnReferenceError,
    NotTransformationReferenceError,
    TransformationApplicationError,
    TransformationParsingError,
    TransformationProcessorError,
    TransformationRegistryError,
)
from bitfount.transformations.parser import TransformationsParser
from bitfount.transformations.processor import TransformationProcessor
from bitfount.transformations.unary_operations import (
    InclusionTransformation,
    OneHotEncodingTransformation,
)

__all__: List[str] = [
    "AdditionTransformation",
    "AlbumentationsImageTransformation",
    "CleanDataTransformation",
    "ComparisonTransformation",
    "DivisionTransformation",
    "IMAGE_TRANSFORMATIONS",
    "InclusionTransformation",
    "IncorrectReferenceError",
    "InvalidBatchTransformationError",
    "MissingColumnReferenceError",
    "MultiplicationTransformation",
    "NormalizeDataTransformation",
    "NotColumnReferenceError",
    "NotTransformationReferenceError",
    "OneHotEncodingTransformation",
    "SubtractionTransformation",
    "TransformationApplicationError",
    "TransformationsParser",
    "TransformationParsingError",
    "TransformationProcessor",
    "TransformationProcessorError",
    "TransformationRegistryError",
    "TRANSFORMATION_REGISTRY",
]

# See top level `__init__.py` for an explanation
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False



"""Callbacks and wrappers for async message service calls."""
from __future__ import annotations

import asyncio
import atexit
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from functools import wraps
import inspect
import threading
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Collection,
    Coroutine,
    Dict,
    Final,
    Generator,
    Generic,
    Optional,
    Set,
    TypeVar,
    Union,
    cast,
)
import uuid
import warnings
import weakref

from bitfount.federated.exceptions import DecryptError
from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _BitfountMessageType,
)

if TYPE_CHECKING:
    from bitfount.federated.transport.base_transport import _BaseMailbox, Handler

from asyncio.futures import Future as AsyncFuture
from concurrent.futures import Future as ConcurrentFuture
from threading import Lock as ThreadingLock

from bitfount.federated.logging import _get_federated_logger

logger = _get_federated_logger(__name__)

_PRIORITY_HANDLER_MAX_WORKERS: Final = 5

# Return type placeholder
_R = TypeVar("_R")


class _AsyncCallback(Generic[_R]):
    """Async wrapper around a callback function.

    Allows us to `await` on the result of this callback. By overriding __call__
    the fact that we've wrapped the callback is transparent to the calling code.
    """

    def __init__(self, fn: Callable[[_BitfountMessage], _R]):
        """Create a new AsyncCallback.

        Args:
            fn: the callback function to be wrapped.
        """
        self._fn = fn
        self._result_exists = asyncio.Event()
        self._result: _R

    def __call__(self, message: _BitfountMessage) -> None:
        """Call the underlying (synchronous) callback function."""
        # Overriding __call__ allows us to transparently wrap the underlying
        # function call so that the call to the async callback looks just like
        # a normal call to the function itself.
        self._result = self._fn(message)
        self._result_exists.set()

    async def result(self, timeout: Optional[int] = None) -> _R:
        """Asynchronously retrieve the result of the callback.

        Will (non-blockingly) wait on the callback to be called.

        Args:
            timeout: Timeout in seconds to await on the result. If not
                provided, will wait indefinitely. Optional.

        Returns:
            The return value of the callback.

        Raises:
            asyncio.TimeoutError: If timeout provided and result is not set within
                timeout seconds.
        """
        if timeout:
            await asyncio.wait_for(self._result_exists.wait(), timeout)
        else:
            await self._result_exists.wait()
        return self._result

    def reset(self) -> None:
        """Clears the result of the callback, allowing it to be re-used."""
        # We don't need to clear the actual result here as that's set before the
        # _result_exists is set.
        self._result_exists.clear()


def _simple_message_returner(x: _BitfountMessage) -> _BitfountMessage:
    """Simple callback that simply returns the message."""
    return x


def _get_message_awaitable() -> _AsyncCallback[_BitfountMessage]:
    """Returns an awaitable wrapper around message retrieval."""
    return _AsyncCallback(_simple_message_returner)


class _AsyncMultipleResponsesHandler:
    """Wraps multiple expected responses in a singular awaitable."""

    def __init__(
        self,
        handler: Handler,
        message_types: Union[_BitfountMessageType, Collection[_BitfountMessageType]],
        mailbox: _BaseMailbox,
        responders: Collection[str],
    ):
        """Creates a handler for multiple responses of a given type(s).

        When expecting multiple separate responses from a set of responders, this
        class will provide an awaitable that returns when either all expected responses
        have been received, or when a timeout is reached (in which case it returns
        the set of those who didn't respond).

        Each message is passed to the assigned handler and track is kept of those
        who have responded. The awaitable returned blocks asynchronously on all
        responses being received.

        Can be used as a context manager which ensures that all message type handlers
        are correctly attached and removed at the end of the usage.

        Args:
            handler: The async function to call for each received message.
            message_types: The message types to handle.
            mailbox: The mailbox where messages will be received.
            responders: The set of expected responders.
        """
        self._orig_handler = handler
        if not isinstance(message_types, Iterable):
            message_types = [message_types]
        self._message_types = message_types
        self._mailbox = mailbox
        self.responders = responders

        # Initialise to the full set of expected and remove them as they response.
        self._not_responded = set(responders)

        # Synchronization primitives for handling multiple responses coming in
        # simultaneously and for keeping track of when all responses have been received.
        self._lock = asyncio.Lock()
        self._responses_done = asyncio.Event()
        self._timeout_reached = False

    async def handler(self, message: _BitfountMessage) -> None:
        """An augmented handler for multiple responses.

        Wraps the supplied handler and tracks the expected responses.

        Args:
            message: The message to be processed.
        """
        # We want to wrap the supplied handler with additional logic to (a) avoid
        # multiple calls to the handler simultaneously which may mess with state,
        # and (b) to enable us to monitor when all responses have been received so
        # we can exit.

        # This lock prevents multiple calls to the handler at the same time
        async with self._lock:
            # This check prevents calls being processed after we have marked it
            # as done, for instance if timeout has occurred.
            if not self._responses_done.is_set():
                # We mark the responder as responded and handle cases where we
                # receive an unexpected response.
                try:
                    self._not_responded.remove(message.sender)
                except KeyError:
                    if message.sender in self.responders:
                        logger.error(
                            f"Received multiple responses from {message.sender}; "
                            f"only expecting one response per responder."
                        )
                    else:
                        logger.error(
                            f"Received unexpected response from {message.sender}; "
                            f"they were not in the list of expected responders."
                        )
                # Once marked as responded we can call the underlying handler and then
                # check whether all responses have been received.
                else:
                    # As this supports both sync and async handlers we need to
                    # process the result (which should be None, but could be a
                    # Coroutine returning None). As such, we comfortably call the
                    # handler and then simply await the result if needed.
                    handler_return = self._orig_handler(message)
                    if inspect.isawaitable(handler_return):
                        # Mypy needs some assurance, despite the above check
                        handler_return = cast(Awaitable[None], handler_return)
                        await handler_return

                    if len(self._not_responded) == 0:
                        self._responses_done.set()
            # Handle responses received after we're marked as done
            else:
                if self._timeout_reached:
                    logger.warning(
                        f"Message received after timeout reached; "
                        f"responder too slow, message will not be processed: {message}"
                    )
                else:
                    logger.warning(
                        f"Received unexpected message after all responses were "
                        f"accounted for: {message}"
                    )

    def __enter__(self) -> _AsyncMultipleResponsesHandler:
        self.setup_handlers()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.remove_handlers()

    def setup_handlers(self) -> None:
        """Setup the augmented handler for the supplied message types."""
        for message_type in self._message_types:
            self._mailbox.register_handler(message_type, self.handler)

    def remove_handlers(self) -> None:
        """Remove the augmented handler for the supplied message types."""
        for message_type in self._message_types:
            self._mailbox.delete_handler(message_type, self.handler)

    async def wait_for_responses(self, timeout: Optional[int] = None) -> Set[str]:
        """Waits for the set of responses to be handled.

        Each response is passed to the supplied (augmented) handler and this method
        will return once all responders have responded or until timeout is reached.

        Args:
            timeout: Optional. Timeout in seconds to wait for all responses to be
                received. If provided, any responders who failed to respond in time
                will be returned as a set.

        Returns:
            The set of responders who did not respond in time.
        """
        # Wait for all responses to have been received or
        # until the timeout expires (if provided).
        try:
            if timeout:
                await asyncio.wait_for(self._responses_done.wait(), timeout)
            else:
                await self._responses_done.wait()
        except asyncio.TimeoutError:
            self._timeout_reached = True
            # Acquiring the lock guarantees that no other responses are
            # _currently_ being processed (and hence we can set the event)
            async with self._lock:
                # Setting this stops other responses in the event loop from
                # being processed _later_.
                self._responses_done.set()

        # Will be empty if all responses received
        return self._not_responded


@dataclass(frozen=True)
class _TemporaryHandler:
    """A wrapper indicating that a handler is to be used only once."""

    handler: Handler


# Create a thread pool for executing high-priority handlers and register its
# shutdown method to be run on program exit.
# Also create a container to store pending futures and to allow us to cancel them
# in the case of a shutdown.
_priority_handler_thread_pool = ThreadPoolExecutor(
    max_workers=_PRIORITY_HANDLER_MAX_WORKERS, thread_name_prefix="priority-handler"
)
_priority_handler_futures: Set[ConcurrentFuture] = set()
_priority_handler_futures_lock: Final = ThreadingLock()


def _shutdown_priority_handler_thread_pool() -> None:
    """Shutdown handler to ensure thread pool shutdown."""
    # Shutdown thread pool (any running tasks may run until completion)
    _priority_handler_thread_pool.shutdown(wait=False)

    # Cancel all outstanding futures
    # NOTE: In Python 3.9+ we can use the built-in ThreadPoolExecutor.shutdown()
    #       method to also cancel futures.
    with _priority_handler_futures_lock:
        for f in _priority_handler_futures:
            f.cancel()


atexit.register(_shutdown_priority_handler_thread_pool)


def _concurrent_future_done(fut: ConcurrentFuture) -> None:
    """Callback function for removing futures from _priority_handler_futures."""
    # As the handler is finished we can remove it from the set of futures that we
    # need to consider for cancellation. We use `discard()` here as we don't care
    # if it's already been removed.
    with _priority_handler_futures_lock:
        _priority_handler_futures.discard(fut)


def _wrap_handler_with_lock(
    fn: Callable[[_BitfountMessage], _R], lock: ThreadingLock
) -> Callable[[_BitfountMessage], _R]:
    """Wraps the target callable in a lock acquisition."""

    @wraps(fn)
    def _wrapped(message: _BitfountMessage) -> _R:
        with lock:
            return fn(message)

    return _wrapped


def _wrap_async_func(
    fn: Callable[[_BitfountMessage], Coroutine[Any, Any, _R]]
) -> Callable[[_BitfountMessage], _R]:
    """Wraps the target async callable in a sync wrapper."""

    @wraps(fn)
    def _wrapped(message: _BitfountMessage) -> _R:
        return asyncio.run(fn(message))

    return _wrapped


class _PriorityHandler(Generic[_R]):
    """A handler that executes with priority by running in a separate thread."""

    def __init__(
        self,
        fn: Union[
            Callable[[_BitfountMessage], _R],
            Callable[[_BitfountMessage], Coroutine[Any, Any, _R]],
        ],
        set_exclusive: bool = True,
    ):
        """Create new priority handler.

        Args:
            fn: The underlying handler to wrap.
            set_exclusive: Whether the handler should only allow one running call
                at any given time.
        """
        self._orig_fn = fn

        self._fn: Callable[[_BitfountMessage], _R]
        if asyncio.iscoroutinefunction(self._orig_fn):
            self._fn = _wrap_async_func(
                cast(
                    Callable[[_BitfountMessage], Coroutine[Any, Any, _R]],
                    self._orig_fn,
                )
            )
        else:
            self._fn = cast(Callable[[_BitfountMessage], _R], self._orig_fn)

        # The lock is managed in here because it needs to be passed into the calling
        # thread and to ensure that only a single call to this handler can be being
        # run at a time. This precludes us from having the lock external to the
        # handler (i.e. in the _HandlerRegister) without having to have a different
        # call signature for the handler (to allow the lock to be passed in).
        self._lock: Optional[ThreadingLock] = None
        if set_exclusive:
            self.set_exclusive()

        # These are used to monitor for the event result and allow us to await on it.
        # NOTE: asyncio.Event() is not thread safe, and so this should only be
        #       accessed from the _calling_ thread. AsyncFuture is not _inherently_
        #       thread safe, but because we create it by wrapping a ConcurrentFuture
        #       it is (as the thread itself interacts with the ConcurrentFuture,
        #       not the AsyncFuture directly).
        self._called = asyncio.Event()
        self._fut: AsyncFuture[_R]

    def __call__(self, message: _BitfountMessage) -> None:
        """Call the underlying handler in a thread."""
        c_fut = _priority_handler_thread_pool.submit(self._fn, message)

        # Register the concurrent future for later shutdown cancellation if needed
        # and adds done_callback
        with _priority_handler_futures_lock:
            _priority_handler_futures.add(c_fut)
        c_fut.add_done_callback(_concurrent_future_done)

        self._fut = asyncio.wrap_future(c_fut)
        self._called.set()

    def set_exclusive(self) -> None:
        """Sets the handler so that only one instance can be running at a time.

        If handler is already marked as exclusive, does nothing.
        """
        if not self._lock:
            self._lock = ThreadingLock()
            self._fn = _wrap_handler_with_lock(self._fn, self._lock)

    @property
    def lock(self) -> Optional[ThreadingLock]:
        """Get the underlying threading.Lock if presented.

        If exclusivity is set, will return the threading.Lock used to ensure this,
        otherwise None.
        """
        return self._lock

    async def _result(self) -> _R:
        """Handles actual result retrieval."""
        # If handler not yet called, wait for that to occur
        await self._called.wait()

        # Then wait for handler to complete
        return await self._fut

    async def result(self, timeout: Optional[int] = None) -> _R:
        """Asynchronously retrieve the result of the callback.

        Will (non-blockingly) wait on the callback to be called.

        Args:
            timeout: Timeout in seconds to await on the result. If not
                provided, will wait indefinitely. Optional.

        Returns:
            The return value of the callback.

        Raises:
            asyncio.TimeoutError: If timeout provided and result is not set within
                timeout seconds.
        """
        return await asyncio.wait_for(self._result(), timeout)

    def __await__(self) -> Generator[Any, None, _R]:
        """Allows `await` functionality directly on the result of the handler."""
        return self.result().__await__()


class _OnlineResponseHandling:
    """A class for monitoring and handling of online checking and responses."""

    def __init__(self, modeller_name: str, aes_key: bytes) -> None:
        """Create a new instance.

        Args:
            modeller_name: The name of the modeller in this task. Used to determine
                which of the received messages should count as online responses.
            aes_key: The encryption key that message bodies will be encrypted with.
        """
        self._modeller_name = modeller_name
        self._aes_key = aes_key
        self._lock = threading.Lock()
        self._waiting: Dict[str, threading.Event] = {}

        # Create a finalizer so that, when this _OnlineResponseHandling instance
        # goes out of scope, any remaining listeners for events are resolved.
        self._finalizer = weakref.finalize(
            self, _OnlineResponseHandling._clear_event_dict, self._waiting, self._lock
        )

    def get_online_check_id(self) -> str:
        """Generate a UUID to place in the message body.

        Also creates an event to mark that we are waiting for a response.
        """
        with self._lock:
            ident = uuid.uuid4().hex
            self._waiting[ident] = threading.Event()
            return ident

    def _mark_all_waiting(self) -> None:
        """Mark all pending events as set and clear them."""
        with self._lock:
            for event in self._waiting.values():
                event.set()
            self._waiting.clear()

    def remove_waiter(self, ident: str) -> None:
        """Remove a specific waiter and mark it as done."""
        with self._lock:
            try:
                event = self._waiting.pop(ident)
                event.set()
            except KeyError:
                # Potentially already removed
                return

    def response_handler(self, message: _BitfountMessage) -> None:
        """Handles messages for online responses.

        Will handle any message type from the intended modeller and use any being
        received to indicate that the modeller is still online.

        Args:
            message: The message to handle.
        """
        if message.sender != self._modeller_name:
            # This isn't relevant for our modeller online checking
            return

        if message.message_type != _BitfountMessageType.ONLINE_RESPONSE:
            # Not an explicit online response but shows modeller _is_ online.
            # Mark all waiting as done.
            self._mark_all_waiting()
        else:  # message.message_type == _BitfountMessageType.ONLINE_RESPONSE
            try:
                decrypted = message.decrypt(self._aes_key)
            except DecryptError:
                logger.warning(
                    "Unable to decrypt ONLINE_RESPONSE message for response checking;"
                    " likely this is not a modeller->pod message."
                )
                # Mark all done as cannot establish _which_ it was intended for
                # but know that modeller is online
                self._mark_all_waiting()
            else:
                if decrypted.body is None:
                    # Older ONLINE_RESPONSE messages will not have the ident
                    # in the body, and so we just have to treat it as though
                    # it is _any_ response
                    warnings.warn(
                        "Support for ID-less ONLINE_RESPONSE messages will be"
                        " removed in a future release, please update your"
                        " bitfount version.",
                        DeprecationWarning,
                    )
                    self._mark_all_waiting()
                else:
                    # Otherwise, try and mark only that specific response monitor
                    try:
                        with self._lock:
                            self._waiting[decrypted.body].set()
                    except KeyError:
                        logger.debug(
                            f"Not waiting on this response ident {decrypted.body}"
                        )

    async def wait_for_response(self, ident: str) -> None:
        """Provides an async interface to wait for modeller online checking."""
        with self._lock:
            try:
                wait_event = self._waiting[ident]
            except KeyError:
                logger.debug(f"Not waiting on this response ident {ident}")
                return None

        # Otherwise wait on the event to be set or for a request to stop waiting
        stop_event = threading.Event()
        loop = asyncio.get_running_loop()
        fut = loop.run_in_executor(
            None, _wait_for_event_with_stop, wait_event, stop_event
        )
        try:
            await fut
        except asyncio.CancelledError as ce:
            # If us waiting on this is cancelled, need to ensure the wait
            # thread finishes as well
            fut.cancel()
            stop_event.set()
            raise ce
        return None

    @staticmethod
    def _clear_event_dict(
        event_dict: Dict[str, threading.Event], lock: threading.Lock
    ) -> None:
        """Clear the event dictionary by setting all events to done.

        This is used by the finalizer to ensure no threads are left waiting. Due
        to this it must be @staticmethod to not have a reference to the
        _OnlineResponseHandling instance (which would block the finalizer from
        running).
        """
        acquired = lock.acquire(blocking=False)

        if not acquired:
            raise RuntimeError(
                "Could not acquire lock to finalise online response handling events"
            )
        else:
            try:
                for event in event_dict.values():
                    event.set()
                event_dict.clear()
            finally:
                lock.release()


def _wait_for_event_with_stop(
    wait_event: threading.Event,
    stop_event: threading.Event,
    polling_timeout: float = 5.0,
) -> None:
    """Helper function that waits on an event but allows exiting early.

    Monitors the wait_event but with a timeout so it can periodically check if it
    should stop early.
    """
    while not stop_event.is_set():
        wait_event.wait(timeout=polling_timeout)
        if wait_event.is_set():
            return



"""Tests for the authentication mechanism."""
from datetime import datetime, timezone
import functools
import json
from pathlib import Path
from typing import Dict, Literal, Optional, Union, cast
from unittest.mock import Mock, PropertyMock, call

from _pytest.capture import CaptureFixture
from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
from dateutil.relativedelta import relativedelta
import jwt
import pytest
from pytest import fixture, raises
from pytest_lazyfixture import lazy_fixture
from pytest_mock import MockerFixture
import responses

from bitfount import web_utils
from bitfount.config import (
    _DEVELOPMENT_ENVIRONMENT,
    _PRODUCTION_ENVIRONMENT,
    _STAGING_ENVIRONMENT,
)
from bitfount.hub.authentication_flow import (
    _AUTHORIZATION_PENDING_ERROR,
    _DEVELOPMENT_AUTH_DOMAIN,
    _DEVELOPMENT_CLIENT_ID,
    _DEVICE_CODE_GRANT_TYPE,
    _PRODUCTION_AUTH_DOMAIN,
    _PRODUCTION_CLIENT_ID,
    _SLOW_DOWN_ERROR,
    _STAGING_AUTH_DOMAIN,
    _STAGING_CLIENT_ID,
    _USERNAME_KEY,
    BitfountSession,
    _AuthEnv,
    _get_auth_environment,
)
from bitfount.hub.types import (
    APIKeys,
    _DeviceAccessTokenRequestDict,
    _DeviceAccessTokenResponseJSON,
    _DeviceCodeRequestDict,
    _DeviceCodeResponseJSON,
    _TokenRefreshRequestDict,
)
from tests.utils.helper import get_warning_logs, unit_test


@fixture
def username() -> str:
    """Username."""
    return "someUsername"


@fixture
def auth_domain() -> str:
    """Authentication domain as fixture."""
    return "some.auth.domain.com"


@fixture
def client_id() -> str:
    """Auth token client ID fixture."""
    return "oljorknkjnio"


@fixture
def scopes() -> str:
    """Auth token scopes fixture."""
    return "spaced out list of scopes"


@fixture
def audience() -> str:
    """Auth token audience fixture."""
    return "some.api.domain.com"


@fixture
def user_storage_path(tmp_path: Path, username: str) -> Path:
    """Temporary directory to act as user's .bitfount directory."""
    user_storage_path = tmp_path / username
    user_storage_path.mkdir()
    return user_storage_path


@fixture
def token_file(user_storage_path: Path) -> Path:
    """Path to save/load tokens from."""
    return user_storage_path / ".token"


@fixture
def bitfount_session(
    audience: str,
    auth_domain: str,
    client_id: str,
    scopes: str,
    user_storage_path: Path,
    username: str,
) -> BitfountSession:
    """Bitfount session instance as fixture."""
    session = BitfountSession(auth_domain, client_id, username, scopes, audience)
    session.user_storage_path = user_storage_path
    session.token_file = user_storage_path / ".token"
    return session


@fixture
def mock_webbrowser(mocker: MockerFixture) -> Mock:
    """Mock webbrowser import."""
    mock_webbrowser: Mock = mocker.patch(
        "bitfount.hub.authentication_flow.webbrowser", autospec=True
    )
    return mock_webbrowser


@fixture
def mock_time(mocker: MockerFixture) -> Mock:
    """Mock time import."""
    mock_time: Mock = mocker.patch(
        "bitfount.hub.authentication_flow.time", autospec=True
    )
    return mock_time


@fixture
def mock_datetime(mocker: MockerFixture) -> Mock:
    """Mock datetime import."""
    mock_datetime: Mock = mocker.patch(
        "bitfount.hub.authentication_flow.datetime", autospec=True
    )
    return mock_datetime


@unit_test
class TestBitfountSession:
    """Tests for the custom BitfountSession."""

    @fixture
    def device_code_request(
        self, audience: str, client_id: str, scopes: str
    ) -> _DeviceCodeRequestDict:
        """Expected request data for /oauth/device/code."""
        return {"audience": audience, "scope": scopes, "client_id": client_id}

    @fixture
    def device_code_response(self) -> _DeviceCodeResponseJSON:
        """Expected response from /oauth/device/code."""
        return {
            "device_code": "Ag_EE...ko1p",
            "user_code": "QTZL-MCBW",
            "verification_uri": "https://accounts.acmetest.org/activate",
            "verification_uri_complete": "https://accounts.acmetest.org"
            "/activate?user_code=QTZL-MCBW",
            "expires_in": 900,
            "interval": 5,
        }

    @fixture
    def device_code(self) -> str:
        """Device code."""
        return "someDeviceCode"

    @fixture
    def access_token(self) -> str:
        """Access token."""
        return "someAccessToken"

    @fixture
    def token_request(
        self, client_id: str, device_code: str
    ) -> _DeviceAccessTokenRequestDict:
        """Expected request data for /oauth/token."""
        return {
            "client_id": client_id,
            "grant_type": _DEVICE_CODE_GRANT_TYPE,
            "device_code": device_code,
        }

    @fixture
    def token_response(self) -> _DeviceAccessTokenResponseJSON:
        """Expected response data from /oauth/token."""
        return {
            "access_token": "eyJz93a...k4laUWw",
            "id_token": "eyJ...0NE",
            "refresh_token": "eyJ...MoQ",
            "scope": "...",
            "expires_in": 86400,
            "token_type": "Bearer",
        }

    @fixture
    def refresh_token(self) -> str:
        """Refresh token."""
        return "someRefreshToken"

    @fixture
    def token_refresh_request(
        self, client_id: str, refresh_token: str
    ) -> _TokenRefreshRequestDict:
        """Expected request data for refresh request to /oauth/token."""
        return {
            "client_id": client_id,
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        }

    @fixture
    def id_token_for_username(self, username: str) -> str:
        """The JWT ID token for username."""
        return jwt.encode({_USERNAME_KEY: username}, "Secret key", algorithm="HS256")

    def test_api_keys_are_read_from_environment_variables(
        self, bitfount_session: BitfountSession, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests that the API keys are read from environment variables."""
        monkeypatch.setenv("BITFOUNT_API_KEY_ID", "someApiKeyId")
        monkeypatch.setenv("BITFOUNT_API_KEY", "someApiKey")
        api_keys = bitfount_session.api_keys
        assert isinstance(api_keys, APIKeys)
        assert api_keys.access_key_id == "someApiKeyId"
        assert api_keys.access_key == "someApiKey"

    def test_api_keys_are_cached(
        self, bitfount_session: BitfountSession, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests that the API keys are cached."""
        monkeypatch.setenv("BITFOUNT_API_KEY_ID", "someApiKeyId")
        monkeypatch.setenv("BITFOUNT_API_KEY", "someApiKey")
        api_keys = bitfount_session.api_keys
        assert isinstance(api_keys, APIKeys)
        assert api_keys.access_key_id == "someApiKeyId"
        assert api_keys.access_key == "someApiKey"

        # Reset the API keys
        monkeypatch.setenv("BITFOUNT_API_KEY_ID", "someApiKeyId2")
        monkeypatch.setenv("BITFOUNT_API_KEY", "someApiKey2")

        # The API keys should be cached (i.e. same as they were before)
        assert api_keys.access_key_id == "someApiKeyId"
        assert api_keys.access_key == "someApiKey"

    def test_both_api_key_environment_variables_are_required(
        self, bitfount_session: BitfountSession, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests that both API key environment variables are required."""
        # We only set the `BITFOUNT_API_KEY_ID` environment variable.
        monkeypatch.setenv("BITFOUNT_API_KEY_ID", "someApiKeyId")
        assert bitfount_session.api_keys is None

    def test_api_keys_property_without_api_key_environment_variables(
        self, bitfount_session: BitfountSession
    ) -> None:
        """Tests that the API keys are None if no environment variables are set."""
        assert bitfount_session.api_keys is None

    @pytest.mark.parametrize(
        "url, expected_output",
        [
            ("https://hub.bitfount.com", True),
            ("https://hub.bitfount.com/api/blah", True),
            ("https://hub.staging.bitfount.com", True),
            ("https://am.hub.bitfount.com", False),
            ("http://hub.bitfount.com", False),  # HTTP is not allowed
        ],
    )
    def test_is_hub_url(
        self, bitfount_session: BitfountSession, expected_output: bool, url: str
    ) -> None:
        """Tests that the is_hub_url method returns the expected output."""
        assert bitfount_session._is_hub_url(url) == expected_output

    @pytest.mark.parametrize(
        "url, expected_output",
        [
            ("https://am.hub.bitfount.com", True),
            ("https://am.hub.bitfount.com/api/blah", True),
            ("https://am.hub.staging.bitfount.com", True),
            ("https://hub.bitfount.com", False),
            ("http://am.hub.bitfount.com", False),  # HTTP is not allowed
        ],
    )
    def test_is_am_url(
        self, bitfount_session: BitfountSession, expected_output: bool, url: str
    ) -> None:
        """Tests that the is_am_url method returns the expected output."""
        assert bitfount_session._is_am_url(url) == expected_output

    def test_token_file_path(
        self,
        audience: str,
        auth_domain: str,
        client_id: str,
        scopes: str,
        username: str,
    ) -> None:
        """Tests that token file is in provided path."""
        session = BitfountSession(auth_domain, client_id, username, scopes, audience)
        # Token file is the path provided with the token file specified
        assert session.token_file == session.user_storage_path / ".token"

    def test_authenticate_with_api_keys(
        self, bitfount_session: BitfountSession, mocker: MockerFixture
    ) -> None:
        """Tests that the authenticate method simply returns if API Keys are present."""
        mock_api_keys = mocker.patch.object(
            BitfountSession,
            "api_keys",
            PropertyMock(return_value=APIKeys("keyID", "key")),
        )
        mock_load_token_from_file = mocker.patch.object(
            bitfount_session, "_load_token_from_file"
        )
        bitfount_session.authenticate()
        mock_api_keys.assert_called_once()
        # The method returns before it gets to the step of loading the token from file
        mock_load_token_from_file.assert_not_called()

    @responses.activate
    def test_fetch_device_code_stores_and_returns_required_values(
        self,
        auth_domain: str,
        bitfount_session: BitfountSession,
        device_code_request: _DeviceCodeRequestDict,
        device_code_response: _DeviceCodeResponseJSON,
    ) -> None:
        """Checks that the needed attributes are stored from device code."""
        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/device/code",
            json=device_code_response,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], device_code_request)
                )
            ],
        )

        user_code, verification_uri = bitfount_session._fetch_device_code()

        assert bitfount_session.device_code == device_code_response.get("device_code")
        assert bitfount_session.device_code_arrival_time is not None
        assert bitfount_session.device_code_expires_in == device_code_response.get(
            "expires_in"
        )
        assert bitfount_session.token_request_interval == device_code_response.get(
            "interval"
        )
        assert user_code == device_code_response["user_code"]
        assert verification_uri == device_code_response["verification_uri_complete"]

    def test_do_verification_opens_browser_correctly(
        self,
        bitfount_session: BitfountSession,
        mock_webbrowser: Mock,
    ) -> None:
        """Checks that request to open browser works."""
        bitfount_session._do_verification("hello", "world")
        mock_webbrowser.open.assert_called_once_with("world")

    @responses.activate
    def test_exchange_device_code_for_token_eventually_receives_tokens(
        self,
        auth_domain: str,
        bitfount_session: BitfountSession,
        device_code: str,
        mock_time: Mock,
        token_request: _DeviceAccessTokenRequestDict,
        token_response: _DeviceAccessTokenResponseJSON,
    ) -> None:
        """Checks that device code successfully exchanged for token."""
        bitfount_session.device_code = device_code
        bitfount_session.device_code_arrival_time = datetime.now(timezone.utc)
        bitfount_session.device_code_expires_in = 2
        bitfount_session.token_request_interval = 1

        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json={"error": _AUTHORIZATION_PENDING_ERROR},
            status=400,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_request)
                )
            ],
        )
        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json=token_response,
            status=200,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_request)
                )
            ],
        )

        bitfount_session._exchange_device_code_for_token()

        assert len(responses.calls) == 2
        assert bitfount_session._access_token == token_response["access_token"]
        assert bitfount_session.refresh_token == token_response["refresh_token"]
        assert bitfount_session.id_token == token_response["id_token"]

        assert bitfount_session.access_token_expires_at is not None
        assert bitfount_session.access_token_expires_at > datetime.now(timezone.utc)
        mock_time.sleep.assert_called_once_with(1)

    @responses.activate
    def test_exchange_device_code_for_token_eventually_receives_error(
        self,
        auth_domain: str,
        bitfount_session: BitfountSession,
        device_code: str,
        mock_time: Mock,
        token_request: _DeviceAccessTokenRequestDict,
    ) -> None:
        """The token endpoint here eventually informs us that the user denied access."""
        bitfount_session.device_code = device_code
        bitfount_session.device_code_arrival_time = datetime.now(timezone.utc)
        bitfount_session.device_code_expires_in = 2
        bitfount_session.token_request_interval = 1

        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json={"error": _AUTHORIZATION_PENDING_ERROR},
            status=400,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_request)
                )
            ],
        )
        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json={"error": "access_denied"},
            status=400,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_request)
                )
            ],
        )

        with raises(ConnectionError):
            bitfount_session._exchange_device_code_for_token()

        assert len(responses.calls) == 2
        assert bitfount_session._access_token is None
        assert bitfount_session.refresh_token is None
        assert bitfount_session.id_token is None
        assert bitfount_session.access_token_expires_at is None
        assert all(map(lambda x: x == call(1), mock_time.sleep.call_args_list))

    @responses.activate
    def test_exchange_device_code_for_token_never_receives_tokens(
        self,
        auth_domain: str,
        bitfount_session: BitfountSession,
        device_code: str,
        mock_datetime: Mock,
        mock_time: Mock,
        token_request: _DeviceAccessTokenRequestDict,
    ) -> None:
        """The token endpoint here never issues tokens, so the device code expires."""
        bitfount_session.device_code = device_code
        bitfount_session.device_code_arrival_time = datetime(
            2020, 1, 14, tzinfo=timezone.utc
        )
        bitfount_session.device_code_expires_in = 300
        bitfount_session.token_request_interval = 1

        mock_datetime.now.side_effect = [
            datetime(2020, 1, 14, minute=x, tzinfo=timezone.utc)
            for x in range(0, 60, 2)
        ]

        auth_server_response = {"error": _AUTHORIZATION_PENDING_ERROR}

        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json=auth_server_response,
            status=400,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_request)
                )
            ],
        )

        with raises(ConnectionError):
            bitfount_session._exchange_device_code_for_token()

        assert len(responses.calls) == 3

        assert all(map(lambda x: x == call(1), mock_time.sleep.call_args_list))
        assert len(mock_time.sleep.call_args_list) == 3

        assert bitfount_session._access_token is None
        assert bitfount_session.refresh_token is None
        assert bitfount_session.id_token is None
        assert bitfount_session.access_token_expires_at is None

    @responses.activate
    def test_exchange_device_code_for_token_receives_slow_down(
        self,
        auth_domain: str,
        bitfount_session: BitfountSession,
        caplog: LogCaptureFixture,
        capsys: CaptureFixture,
        device_code: str,
        mock_time: Mock,
        token_request: _DeviceAccessTokenRequestDict,
        token_response: _DeviceAccessTokenResponseJSON,
    ) -> None:
        """Test interval increased if receive slow down response."""
        # Set relevant attributes on session
        bitfount_session.device_code = device_code
        bitfount_session.device_code_arrival_time = datetime.now(timezone.utc)
        bitfount_session.device_code_expires_in = 300
        bitfount_session.token_request_interval = 1

        # First and second responses are "slow down"
        for _ in range(2):
            responses.add(
                responses.POST,
                f"https://{auth_domain}/oauth/token",
                json={"error": _SLOW_DOWN_ERROR},
                status=400,
                match=[
                    responses.matchers.urlencoded_params_matcher(
                        cast(Dict[str, str], token_request)
                    )
                ],
            )
        # Final response is correct response
        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json=token_response,
            status=200,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_request)
                )
            ],
        )

        bitfount_session._exchange_device_code_for_token()

        # Check warning logs for interval notes
        warning_logs = get_warning_logs(caplog)
        assert (
            "Polling too quickly; increasing interval from 1 to 2 seconds"
            in warning_logs
        )
        assert (
            "Polling too quickly; increasing interval from 2 to 3 seconds"
            in warning_logs
        )
        # Check stdout for increased interval waits
        stdout = capsys.readouterr().out
        assert (
            "Awaiting authentication in browser. Will check again in 1 seconds."
            not in stdout
        )
        assert (
            "Awaiting authentication in browser. Will check again in 2 seconds."
            in stdout
        )
        assert (
            "Awaiting authentication in browser. Will check again in 3 seconds."
            in stdout
        )

        # Check correct number of requests made
        assert len(responses.calls) == 3

        # Check sleep called with updated intervals
        assert mock_time.sleep.call_args_list == [call(2), call(3)]

    @pytest.mark.parametrize(
        argnames=("content_type", "content_value", "expected_error_msg", "in_stdout"),
        argvalues=(
            pytest.param(
                "json",
                {},
                'An unexpected error occurred: status code: 404; "{}"',
                True,
                id="incorrect json",
            ),
            pytest.param(
                "body",
                "error but not necessarily json",
                (
                    "Received 404 status response, but JSON is invalid: "
                    '"error but not necessarily json"'
                ),
                False,
                id="non-json response",
            ),
        ),
    )
    @responses.activate
    def test_exchange_device_code_for_token_receives_non_400_error(
        self,
        auth_domain: str,
        bitfount_session: BitfountSession,
        caplog: LogCaptureFixture,
        capsys: CaptureFixture,
        content_type: Literal["json", "body"],
        content_value: Union[dict, str],
        device_code: str,
        expected_error_msg: str,
        in_stdout: bool,
        mock_time: Mock,
        remove_web_retry_backoff_sleep: Optional[Mock],
        token_request: _DeviceAccessTokenRequestDict,
        token_response: _DeviceAccessTokenResponseJSON,
    ) -> None:
        """Test error raised if receive non-400 error response."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        # Set relevant attributes on session
        bitfount_session.device_code = device_code
        bitfount_session.device_code_arrival_time = datetime.now(timezone.utc)
        bitfount_session.device_code_expires_in = 300
        bitfount_session.token_request_interval = 1

        partial_response_add = functools.partial(
            responses.add,
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            status=404,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_request)
                )
            ],
        )
        if content_type == "json":
            partial_response_add(json=content_value)
        else:
            partial_response_add(body=content_value)

        with pytest.raises(
            ConnectionError,
            match="Failed to retrieve a device code from the authentication server",
        ):
            bitfount_session._exchange_device_code_for_token()

        # Check logger/stdout for errors
        assert expected_error_msg in caplog.text
        if in_stdout:
            assert expected_error_msg in capsys.readouterr().out

        # Check retries occurred
        assert (
            remove_web_retry_backoff_sleep.call_count == web_utils._DEFAULT_MAX_RETRIES
        )

    @responses.activate
    def test_bitfount_session_request_provides_token(
        self,
        access_token: str,
        bitfount_session: BitfountSession,
        device_code: str,
    ) -> None:
        """Checks that `request()` call provides token."""
        bitfount_session._access_token = access_token
        bitfount_session.device_code = device_code
        # Expires 1 month in future
        bitfount_session.access_token_expires_at = datetime.now(
            timezone.utc
        ) + relativedelta(months=1)

        api_url = "https://some.api.url/goes/here"

        responses.add(
            responses.POST,
            api_url,
        )

        # Send request
        bitfount_session.request("POST", api_url)

        # Check request had access token
        assert (
            responses.calls[0].request.headers["authorization"]
            == f"Bearer {access_token}"
        )

    @responses.activate
    def test_bitfount_session_request_refreshes_expired_then_provides_token(
        self,
        access_token: str,
        auth_domain: str,
        bitfount_session: BitfountSession,
        device_code: str,
        id_token_for_username: str,
        mock_time: Mock,
        refresh_token: str,
        token_file: Path,
        token_refresh_request: _TokenRefreshRequestDict,
        token_response: _DeviceAccessTokenResponseJSON,
    ) -> None:
        """Checks that if token expired new one is provided and saved."""
        bitfount_session._access_token = access_token
        bitfount_session.device_code = device_code
        bitfount_session.refresh_token = refresh_token
        bitfount_session.id_token = id_token_for_username

        # Expire token
        bitfount_session.access_token_expires_at = datetime(
            2000, 1, 14, tzinfo=timezone.utc
        )
        bitfount_session.token_request_interval = 1

        api_url = "https://some.api.url/goes/here"

        # Set expected ID Token on response
        token_response["id_token"] = id_token_for_username

        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            # Token refresh has same form as initial token response so we just reuse it
            json=token_response,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_refresh_request)
                )
            ],
        )

        responses.add(
            responses.POST,
            api_url,
        )

        # Send request
        bitfount_session.request("POST", api_url)

        # Check request had refreshed access token
        assert (
            responses.calls[1].request.headers["authorization"]
            == f"Bearer {token_response['access_token']}"
        )

        # Check refreshed token was saved
        saved_token = json.load(token_file.open())
        assert saved_token["access_token"] == token_response["access_token"]
        assert saved_token["refresh_token"] == token_response["refresh_token"]
        assert saved_token["id_token"] == token_response["id_token"]
        assert datetime.fromtimestamp(
            saved_token["access_token_expires_at"], tz=timezone.utc
        ) > datetime.now(timezone.utc)

        # Check session metadata was saved
        assert saved_token["auth_domain"] == bitfount_session._auth_domain
        assert saved_token["client_id"] == bitfount_session._client_id
        assert saved_token["scopes"] == bitfount_session._scopes
        assert saved_token["audience"] == bitfount_session._audience

        mock_time.sleep.assert_not_called()

    @responses.activate
    def test_bitfount_session_request_forces_manual_login_when_refresh_fails(
        self,
        access_token: str,
        auth_domain: str,
        bitfount_session: BitfountSession,
        device_code: str,
        device_code_request: _DeviceCodeRequestDict,
        device_code_response: _DeviceCodeResponseJSON,
        id_token_for_username: str,
        mock_time: Mock,
        mock_webbrowser: Mock,
        refresh_token: str,
        scopes: str,
        token_refresh_request: _TokenRefreshRequestDict,
        token_request: _DeviceAccessTokenRequestDict,
        token_response: _DeviceAccessTokenResponseJSON,
    ) -> None:
        """Checks that the user is made to log in if the token refresh fails.

        Once they have logged in the web request is completed
        """
        bitfount_session._access_token = access_token
        bitfount_session.device_code = device_code
        bitfount_session.refresh_token = refresh_token
        bitfount_session.id_token = id_token_for_username

        # Expire token
        bitfount_session.access_token_expires_at = datetime(
            2000, 1, 14, tzinfo=timezone.utc
        )
        bitfount_session.token_request_interval = 1

        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json={"error": "some error"},
            status=401,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_refresh_request)
                )
            ],
        )

        # Set new device code on /oauth/device/code response and expected request
        # against /oauth/token
        new_device_code = "newDeviceCode"
        device_code_response["device_code"] = new_device_code
        token_request["device_code"] = new_device_code

        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/device/code",
            json=device_code_response,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], device_code_request)
                )
            ],
        )

        # Set ID Token on token response
        token_response["id_token"] = id_token_for_username

        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json=token_response,
            status=200,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_request)
                )
            ],
        )

        expected_json = {"request": "success"}
        api_url = "https://some.api.url/goes/here"
        responses.add(responses.POST, api_url, json=expected_json)

        # Send request
        response = bitfount_session.request("POST", api_url)

        assert response.json() == expected_json
        mock_time.sleep.assert_called_once()
        mock_webbrowser.open.assert_called_once()

    def test_bitfount_session_authenticates_when_no_token_loaded(
        self,
        bitfount_session: BitfountSession,
        mocker: MockerFixture,
        username: str,
    ) -> None:
        """Checks that token requested and saved when none exists to load."""
        mock_fetch_device_code = mocker.patch.object(
            bitfount_session, "_fetch_device_code", return_value=(1, 2)
        )
        mock_do_verification = mocker.patch.object(bitfount_session, "_do_verification")
        mock_exchange_device_code = mocker.patch.object(
            bitfount_session, "_exchange_device_code_for_token"
        )
        mock_save_token = mocker.patch.object(bitfount_session, "_save_token_to_file")
        mocker.patch.object(
            bitfount_session, "_get_username_from_id_token", return_value=username
        )

        bitfount_session.authenticate()

        mock_fetch_device_code.assert_called_once()
        mock_do_verification.assert_called_once_with(1, 2)
        mock_exchange_device_code.assert_called_once()
        mock_save_token.assert_called_once_with(bitfount_session.token_file)

    @pytest.mark.parametrize(
        "url",
        [
            "hub",  # Hub URL
            "am",  # AM URL
            "other",  # Other URL
        ],
    )
    @responses.activate
    def test_bitfount_session_request_with_api_keys(
        self,
        access_token: str,
        bitfount_session: BitfountSession,
        device_code: str,
        mocker: MockerFixture,
        url: str,
    ) -> None:
        """Checks that `request()` call adds api keys only to Hub and AM calls."""
        # Set API keys and token
        bitfount_session.api_keys = APIKeys(
            "accessKeyID1:accessKeyID2", "accessKey1:accessKey2"
        )
        bitfount_session._access_token = access_token
        bitfount_session.device_code = device_code
        # Expires 1 month in future
        bitfount_session.access_token_expires_at = datetime.now(
            timezone.utc
        ) + relativedelta(months=1)

        # Mock `_is_hub_url` and `_is_am_url` methods
        if url == "hub":
            mocker.patch.object(bitfount_session, "_is_hub_url", return_value=True)
            api_url = "https://this.is.a.hub.url"
        elif url == "am":
            mocker.patch.object(bitfount_session, "_is_am_url", return_value=True)
            api_url = "https://this.is.an.am.url"
        else:
            api_url = "https://this.is.a.different.url"

        # Set up response
        responses.add(
            responses.POST,
            api_url,
        )

        # Send request
        bitfount_session.request("POST", api_url)

        # Check request had appropriate headers
        if url == "hub":
            assert "authorization" not in responses.calls[0].request.headers
            # Check that the request only included the first part of the api keys
            assert responses.calls[0].request.headers["x-api-key-id"] == "accessKeyID1"
            assert responses.calls[0].request.headers["x-api-key"] == "accessKey1"
        elif url == "am":
            assert "authorization" not in responses.calls[0].request.headers
            # Check that the request included the full api keys
            assert (
                responses.calls[0].request.headers["x-api-key-id"]
                == "accessKeyID1:accessKeyID2"
            )
            assert (
                responses.calls[0].request.headers["x-api-key"]
                == "accessKey1:accessKey2"
            )
        else:
            # Confirm that API keys are not used if the URL is not a Hub or AM URL
            assert "x-api-key-id" not in responses.calls[0].request.headers
            assert "x-api-key" not in responses.calls[0].request.headers
            assert (
                responses.calls[0].request.headers["authorization"]
                == f"Bearer {access_token}"
            )

    def test_bitfount_session_authenticates_when_token_loaded_but_expired(
        self,
        access_token: str,
        bitfount_session: BitfountSession,
        id_token_for_username: str,
        mocker: MockerFixture,
        refresh_token: str,
        token_file: Path,
    ) -> None:
        """Checks that a new token is fetched via the login flow.

        This is the case where the loaded token has expired completely.
        """
        # Create saved token
        bitfount_session._access_token = access_token
        bitfount_session.access_token_expires_at = datetime(
            year=2000, month=6, day=13, tzinfo=timezone.utc
        )
        bitfount_session.id_token = id_token_for_username
        bitfount_session.refresh_token = refresh_token
        bitfount_session._save_token_to_file(token_file)

        # Reset session to initial state
        bitfount_session._access_token = None
        bitfount_session.access_token_expires_at = None
        bitfount_session.id_token = None
        bitfount_session.refresh_token = None

        mocker.patch.object(
            bitfount_session, "_refresh_access_token", return_value=False
        )
        mock_fetch_device_code = mocker.patch.object(
            bitfount_session, "_fetch_device_code", return_value=(1, 2)
        )
        mock_do_verification = mocker.patch.object(bitfount_session, "_do_verification")
        mock_exchange_device_code = mocker.patch.object(
            bitfount_session, "_exchange_device_code_for_token"
        )

        bitfount_session.authenticate()

        mock_fetch_device_code.assert_called_once()
        mock_do_verification.assert_called_once_with(1, 2)
        mock_exchange_device_code.assert_called_once()

    def test_bitfount_session_refreshes_when_token_loaded_but_expired(
        self,
        access_token: str,
        bitfount_session: BitfountSession,
        id_token_for_username: str,
        mocker: MockerFixture,
        refresh_token: str,
        token_file: Path,
    ) -> None:
        """Checks that a new token is fetched using the refresh mechanism.

        This is for when the loaded token has expired,
        but the refresh token is still valid
        """
        # Create saved token
        bitfount_session._access_token = access_token
        bitfount_session.access_token_expires_at = datetime(
            year=2000, month=6, day=13, tzinfo=timezone.utc
        )
        bitfount_session.id_token = id_token_for_username
        bitfount_session.refresh_token = refresh_token
        bitfount_session._save_token_to_file(token_file)

        # Reset session to initial state
        bitfount_session._access_token = None
        bitfount_session.access_token_expires_at = None
        bitfount_session.id_token = None
        bitfount_session.refresh_token = None

        mock_refresh_access_token = mocker.patch.object(
            bitfount_session, "_refresh_access_token"
        )
        mock_fetch_device_code = mocker.patch.object(
            bitfount_session, "_fetch_device_code"
        )
        mock_do_verification = mocker.patch.object(bitfount_session, "_do_verification")
        mock_exchange_device_code = mocker.patch.object(
            bitfount_session, "_exchange_device_code_for_token"
        )

        bitfount_session.authenticate()

        # Ensure refresh was called
        mock_refresh_access_token.assert_called_once()
        # Ensure manual login flow is not run
        mock_fetch_device_code.assert_not_called()
        mock_do_verification.assert_not_called()
        mock_exchange_device_code.assert_not_called()

    def test_bitfount_session_authenticated_when_valid_token_loaded(
        self,
        access_token: str,
        bitfount_session: BitfountSession,
        id_token_for_username: str,
        mocker: MockerFixture,
        refresh_token: str,
        token_file: Path,
    ) -> None:
        """Checks that authentication is successful when token loaded from file."""
        # Create saved token
        bitfount_session._access_token = access_token
        # Timestamp for 1 month from now
        bitfount_session.access_token_expires_at = datetime.now(
            timezone.utc
        ) + relativedelta(months=1)
        bitfount_session.id_token = id_token_for_username
        bitfount_session.refresh_token = refresh_token
        bitfount_session._save_token_to_file(token_file)

        # Reset session to initial state
        bitfount_session._access_token = None
        bitfount_session.access_token_expires_at = None
        bitfount_session.id_token = None
        bitfount_session.refresh_token = None

        mock_fetch_device_code = mocker.patch.object(
            bitfount_session, "_fetch_device_code"
        )
        mock_exchange_device_code = mocker.patch.object(
            bitfount_session, "_exchange_device_code_for_token"
        )

        bitfount_session.authenticate()

        mock_fetch_device_code.assert_not_called()
        mock_exchange_device_code.assert_not_called()

    def test_bitfount_session_doesnt_use_loaded_token_when_metadata_differs(
        self,
        access_token: str,
        bitfount_session: BitfountSession,
        id_token_for_username: str,
        mocker: MockerFixture,
        refresh_token: str,
        token_file: Path,
        user_storage_path: Path,
    ) -> None:
        """Test bitfount session will create & save a new token when different."""
        # Create saved token
        old_session_configuration = BitfountSession(
            "someOtherDomain.com",
            "not our client id",
            "someUsername",
            "not our scopes",
            "someOtherAudience",
        )
        old_session_configuration.user_storage_path = user_storage_path
        old_session_configuration._access_token = access_token
        # Timestamp for 1 month from now
        old_session_configuration.access_token_expires_at = datetime.now(
            timezone.utc
        ) + relativedelta(months=1)
        bitfount_session.id_token = id_token_for_username
        old_session_configuration.refresh_token = refresh_token
        old_session_configuration._save_token_to_file(token_file)

        mock_fetch_device_code = mocker.patch.object(
            bitfount_session, "_fetch_device_code", return_value=(1, 2)
        )
        mock_do_verification = mocker.patch.object(bitfount_session, "_do_verification")
        mock_exchange_device_code = mocker.patch.object(
            bitfount_session, "_exchange_device_code_for_token"
        )
        mock_save_token = mocker.patch.object(bitfount_session, "_save_token_to_file")

        bitfount_session.authenticate()

        mock_fetch_device_code.assert_called_once()
        mock_do_verification.assert_called_once_with(1, 2)
        mock_exchange_device_code.assert_called_once()
        mock_save_token.assert_called_once_with(token_file)

    @responses.activate
    def test_send_token_request_to_refresh_expired_token(
        self,
        auth_domain: str,
        bitfount_session: BitfountSession,
        refresh_token: str,
        token_refresh_request: _TokenRefreshRequestDict,
    ) -> None:
        """Checks _send_token_request works with refresh=True."""
        bitfount_session.refresh_token = refresh_token

        auth_server_response = {"some": "response"}

        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json=auth_server_response,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_refresh_request)
                )
            ],
        )

        response = bitfount_session._send_token_request(refresh=True)

        assert response.json() == auth_server_response

    @responses.activate
    def test_send_token_request_as_device_code(
        self,
        auth_domain: str,
        bitfount_session: BitfountSession,
        client_id: str,
        device_code: str,
        token_request: _DeviceAccessTokenRequestDict,
    ) -> None:
        """Checks that _send_token_request works correctly."""
        bitfount_session.device_code = device_code

        auth_server_response = {"some": "response"}

        responses.add(
            responses.POST,
            f"https://{auth_domain}/oauth/token",
            json=auth_server_response,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], token_request)
                )
            ],
        )

        response = bitfount_session._send_token_request()

        assert response.json() == auth_server_response

    def test_get_username_from_id_token(
        self,
        bitfount_session: BitfountSession,
        id_token_for_username: str,
        username: str,
    ) -> None:
        """Tests username can be extracted from ID Token."""
        bitfount_session.id_token = id_token_for_username

        assert bitfount_session._get_username_from_id_token() == username

    @pytest.mark.parametrize(
        "_username, assertion_error",
        [(lazy_fixture("username"), False), ("not-real-username", True)],
    )
    def test_verify_user_storage_path(
        self,
        _username: str,
        assertion_error: bool,
        bitfount_session: BitfountSession,
        mocker: MockerFixture,
    ) -> None:
        """Tests that token won't be saved in a directory with a different username."""
        mocker.patch.object(
            BitfountSession, "username", PropertyMock(return_value=_username)
        )

        if assertion_error:
            with pytest.raises(AssertionError):
                bitfount_session._verify_user_storage_path()
        else:
            bitfount_session._verify_user_storage_path()


@unit_test
@pytest.mark.parametrize(
    argnames=(
        "environment",
        "expected_name",
        "expected_auth_domain",
        "expected_client_id",
    ),
    argvalues=(
        pytest.param(
            _STAGING_ENVIRONMENT,
            "staging",
            _STAGING_AUTH_DOMAIN,
            _STAGING_CLIENT_ID,
            id="staging",
        ),
        pytest.param(
            _DEVELOPMENT_ENVIRONMENT,
            "development",
            _DEVELOPMENT_AUTH_DOMAIN,
            _DEVELOPMENT_CLIENT_ID,
            id="development",
        ),
        pytest.param(
            _PRODUCTION_ENVIRONMENT,
            "production",
            _PRODUCTION_AUTH_DOMAIN,
            _PRODUCTION_CLIENT_ID,
            id="production",
        ),
    ),
)
def test_get_auth_environment(
    environment: str,
    expected_auth_domain: str,
    expected_client_id: str,
    expected_name: str,
    monkeypatch: MonkeyPatch,
) -> None:
    """Tests _get_auth_environment with various envvar values."""
    # Patch out environment variable
    monkeypatch.setenv("BITFOUNT_ENVIRONMENT", environment)

    # Check return
    assert _get_auth_environment() == _AuthEnv(
        expected_name, expected_auth_domain, expected_client_id
    )



"""Tests for the federated model training algorithm."""
from typing import TYPE_CHECKING
from unittest.mock import Mock, create_autospec

import numpy as np
from pytest_mock import MockerFixture

from bitfount import BitfountSchema
from bitfount.backends.pytorch.models.models import PyTorchTabularClassifier
from bitfount.federated.algorithms.base import (
    _BaseAlgorithm,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModelAlgorithm,
    _BaseModellerModelAlgorithm,
    _BaseWorkerModelAlgorithm,
)
from bitfount.federated.algorithms.model_algorithms.federated_training import (
    FederatedModelTraining,
    _BaseModelTrainingMixIn,
    _ModellerSide,
    _WorkerSide,
)
from bitfount.federated.utils import _ALGORITHMS
from bitfount.hub import BitfountHub
from bitfount.schemas.utils import bf_dump, bf_load
from tests.bitfount.backends.pytorch.models.test_models import assert_vars_equal
from tests.utils.helper import backend_test, create_datastructure, unit_test


class TestFederatedModelTraining:
    """Test Federated Model Training algorithm."""

    @unit_test
    def test_modeller(self, model: Mock) -> None:
        """Test modeller method."""
        algorithm_factory = FederatedModelTraining(model=model)
        algorithm = algorithm_factory.modeller()
        for type_ in [
            _BaseAlgorithm,
            _BaseModellerAlgorithm,
            _BaseModelAlgorithm,
            _BaseModellerModelAlgorithm,
        ]:
            assert isinstance(algorithm, type_)
        assert algorithm.steps == 10

    @unit_test
    def test_worker(self, model: Mock) -> None:
        """Test worker method."""
        algorithm_factory = FederatedModelTraining(model=model)
        algorithm = algorithm_factory.worker(
            hub=create_autospec(BitfountHub, instance=True)
        )
        for type_ in [
            _BaseAlgorithm,
            _BaseWorkerAlgorithm,
            _BaseModelAlgorithm,
            _BaseWorkerModelAlgorithm,
        ]:
            assert isinstance(algorithm, type_)
        assert algorithm.steps == 10

    @unit_test
    def test_modeller_logs_validation_metrics(
        self, mocker: MockerFixture, model: Mock
    ) -> None:
        """Test that the validation metrics are logged for the modeller."""
        mocker.patch.object(
            _BaseModelTrainingMixIn,
            "get_param_states",
            return_value={"layer1": np.ndarray([0, 1, 2, 3])},
        )
        mock_model_log = Mock()
        mocker.patch.object(model, "log_", mock_model_log)

        algorithm_factory = FederatedModelTraining(model=model)
        algorithm = algorithm_factory.modeller()

        algorithm.run(validation_metrics={"AUC": 0.8})
        mock_model_log.assert_called()

    @unit_test
    def test_worker_run_calls_diff_params(
        self, mocker: MockerFixture, model: Mock
    ) -> None:
        """Test worker run method calls diff_params.

        Test we return the difference between the old and new model parameters.
        """
        algorithm_factory = FederatedModelTraining(model=model)
        algorithm = algorithm_factory.worker(
            hub=create_autospec(BitfountHub, instance=True)
        )
        mock_diff_params = mocker.patch.object(algorithm, "diff_params")
        data = Mock()
        serialized_model_params = Mock()
        algorithm.run(data, serialized_model_params, iterations=1)
        mock_diff_params.assert_called_once()


@backend_test
@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for column average algorithm."""

    def test_serialization(self) -> None:
        """Test Marshmallow Serialization for column average algorithm."""
        model = PyTorchTabularClassifier(
            datastructure=create_datastructure(),
            schema=BitfountSchema(),
            epochs=2,
        )
        algorithm_factory = FederatedModelTraining(model=model)
        dumped = bf_dump(algorithm_factory)
        loaded = bf_load(dumped, _ALGORITHMS)

        assert algorithm_factory.class_name == loaded.class_name
        assert_vars_equal(vars(algorithm_factory.model), vars(loaded.model))


# Static tests for algorithm-protocol compatibility
if TYPE_CHECKING:
    from typing import cast

    from bitfount.federated.protocols.model_protocols.federated_averaging import (
        _FederatedAveragingCompatibleAlgoFactory,
        _FederatedAveragingCompatibleModeller,
        _FederatedAveragingCompatibleWorker,
    )
    from bitfount.types import (
        DistributedModelProtocol,
        _DistributedModelTypeOrReference,
    )

    # Check compatible with FederatedAveraging
    _algo_factory: _FederatedAveragingCompatibleAlgoFactory = FederatedModelTraining(
        model=cast(_DistributedModelTypeOrReference, object())
    )
    _modeller_side: _FederatedAveragingCompatibleModeller = _ModellerSide(
        model=cast(DistributedModelProtocol, object())
    )
    _worker_side: _FederatedAveragingCompatibleWorker = _WorkerSide(
        model=cast(DistributedModelProtocol, object())
    )
